__author__ = 'peter'
import numpy as np


def split_data_by_label(data, labels, frac_training = 0.5):
    """
    Split the data so that each label gets approximately the correct proportions between the training and test sets
    :param data: An (n_samples, ...) array of data
    :param labels: An (n_samples, ) array of labels
    :param frac_training: The fraction of data to put in the training set vs the test.
    :return: (x_tr, x_ts, y_tr, y_ts)
    """
    # TODO: Write test - it's important that this be correct

    assert len(data)==len(labels)
    labels = np.array(labels) if not isinstance(labels, np.ndarray) else labels
    unique_labels, inverse_ixs = np.unique(labels, return_inverse = True)
    ixs = np.arange(len(data))
    label_indices = [ixs[inverse_ixs==l] for l in xrange(len(unique_labels))]
    cutoffs = [int(np.round(frac_training*len(ixs))) for ixs in label_indices]
    training_indices = np.sort(np.concatenate([ixs[:c] for ixs, c in zip(label_indices, cutoffs)]))
    test_indices = np.sort(np.concatenate([ixs[c:] for ixs, c in zip(label_indices, cutoffs)]))
    return data[training_indices], labels[training_indices], data[test_indices], labels[test_indices]



"""
Using this module, you can turn your main function into an "Experiment", which, when run, stores all console output, plots,
and computed results to disk (in ~/.artemis/experiments)

For a simple demo, see artemis/fileman/demo_experiments.py

Any function that can be called alone with no arguments can be turned into an experiment using the @experiment_function
decorator:

    @experiment_function
    def my_experiment(a=1, b=2, c=3):
        ...

This turns the function into an Experiment object, which has the following methods:

    result = my_experiment.run()
        Run the function, and save all text outputs and plots to disk.
        Console output, plots, and result are stored in: ~/.artemis/experiments

    ex=my_experiment.add_variant(name, **kwargs)
        Add a variant to the experiment, with new arguments that override any existing ones.
        The variant is itself an experiment, and can have variants of its own.

    ex=my_experiment.get_variant(name, *sub_names)
        Get a variant of an experiment (or one of its sub-variants).

To open up a menu where you can see and run all experiments (and their variants) that have been created run:

    browse_experiments()

To browse through records of experiments that have been run, either run this file, or call the function:

    browse_experiment_records()

You may not want an experiment to be runnable (and show up in browse_experiments) - you may just want it
to serve as a basis on which to make variants.  In this case, you can decorate with

    @experiment_root_function
    def my_experiment_root(a=1, b=2, c=3):
        ...

If you want even the variants of this experiment to be roots for other experiments, you can use method

    ex=my_experiment.add_root_variant(name, **kwargs)

To run your experiment (if not a root) and all non-root variants (and sub variants, and so on) of an experiment, run:

    my_experiment.run_all()

If your experiment takes a long time to run, you may not want to run it every time you want to see the plots of
results (this is especially so if you are refining your plotting function, and don't want to run from scratch just to
check your visualization methods).  in this case, you can do the following:

    def my_display_function(data_to_display):
        ...  # Plot data

    @ExperimentFunction(display_function = my_display_function)
    def my_experiment(a=1, b=2, c=3):
        ...
        return data_to_display

Now, you can use the method:

    my_experiment.display_last()
        Load the last results for this experiment (if any) and plot them again.

"""

import atexit
import inspect
import logging
import multiprocessing
import os
import pickle
import re
import shutil
import tempfile
import time
import traceback
from collections import OrderedDict
from contextlib import contextmanager
from datetime import datetime
from functools import partial
from getpass import getuser
from pprint import pprint
from artemis.fileman.local_dir import format_filename, make_file_dir, get_local_path, make_dir
from artemis.fileman.persistent_ordered_dict import PersistentOrderedDict
from artemis.general.display import CaptureStdOut
from artemis.general.functional import infer_derived_arg_values, get_partial_chain
from artemis.general.hashing import compute_fixed_hash
from artemis.general.should_be_builtins import separate_common_items, izip_equal
from artemis.general.test_mode import is_test_mode, set_test_mode
from uuid import getnode
import numpy as np
try:
    from enum import Enum
except ImportError:
    raise ImportError("Failed to import the enum package. This was added in python 3.4 but backported back to 2.4.  To install, run 'pip install --upgrade pip enum34'")

logging.basicConfig()
ARTEMIS_LOGGER = logging.getLogger('artemis')
ARTEMIS_LOGGER.setLevel(logging.INFO)

__author__ = 'peter'


def experiment_function(f):
    """
    Use this decorator (@experiment_function) on a function that you want to run.  e.g.

        @experiment_function
        def demo_my_experiment(a=1, b=2, c=3):
            ...

    To run the experiment in a mode where it records all print statements and matplotlib figures, go:
    demo_my_experiment.run()

    To create a variant on the experiment, with different arguments, go:

        v1 = demo_my_experiment.add_variant('try_a=2', a=2)

    The variant v1 is then also an experiment, so you can go

        v1.run()

    """
    return ExperimentFunction()(f)


def experiment_root(f):
    """
    Use this decorator on a function that you want to build variants off of:

        @experiment_root
        def demo_my_experiment(a, b=2, c=3):
            ...

        demo_my_experiment.add_variant('a1', a=1)
        demo_my_experiment.add_variant('a2', a=2)

    The root experiment is not runnable by itself, and will not appear in the list when you
    browse experiments.
    """
    return ExperimentFunction(is_root=True)(f)


class ExperimentFunction(object):
    """ Decorator for an experiment
    """

    def __init__(self, display_function=None, comparison_function=None, one_liner_results=None, info=None, is_root=False):
        """
        :param display_function: A function that takes the results (whatever your experiment returns) and displays them.
        :param comparison_function: A function that takes an OrderedDict<experiment_name, experiment_return_value>.
            You can optionally define this function to compare the results of different experiments.
            You can use call this via the UI with the compare_results command.
        :param one_liner_results: A function that takes your results and returns a 1 line string summarizing them.
        :param info: Don't use this?
        :param is_root: True to make this a root experiment - so that it is not listed to be run itself.
        """
        self.display_function = display_function
        self.comparison_function = comparison_function
        self.info = info
        self.is_root = is_root
        self.one_liner_results = one_liner_results

    def __call__(self, f):
        f.is_base_experiment = True
        ex = Experiment(
            name=f.__name__,
            function=f,
            display_function=self.display_function,
            comparison_function = self.comparison_function,
            one_liner_results=self.one_liner_results,
            is_root=self.is_root
        )
        return ex


GLOBAL_EXPERIMENT_LIBRARY = OrderedDict()


class ExpInfoFields(Enum):
    NAME = 'Name'
    ID = 'Identifier'
    DIR = 'Directory'
    TIMESTAMP = 'Timestamp'
    ARGS = 'Args'
    FUNCTION = 'Function'
    MODULE = 'Module'
    FILE = 'File'
    STATUS = 'Status'
    N_FIGS = '# Figures Generated'
    FIGS = 'Figure Locs'
    RUNTIME = 'Run Time'
    VERSION = 'Version'
    NOTES = 'Notes'
    USER = 'User'
    MAC = 'MAC Address'


class ExpStatusOptions(Enum):
    STARTED = 'Running (or Killed)'
    ERROR = 'Error'
    STOPPED = 'Stopped by User'
    FINISHED = 'Ran Succesfully'


class ExperimentRecordInfo(object):
    def __init__(self, file_path, write_text_version=True):
        before, ext = os.path.splitext(file_path)
        assert ext == '.pkl', 'Your file-path must be a pickle'
        self._text_path = before + '.txt' if write_text_version else None
        self.persistent_obj = PersistentOrderedDict(file_path=file_path)

    def has_field(self, field):
        assert field in ExpInfoFields, 'Field must be a member of ExperimentRecordInfo.FIELDS'
        return field in self.persistent_obj

    def get_field(self, field):
        """
        :param field: A member of ExperimentRecordInfo.FIELDS
        :return:
        """
        assert field in ExpInfoFields, 'Field must be a member of ExperimentRecordInfo.FIELDS'
        return self.persistent_obj[field]

    def set_field(self, field, value):
        assert field in ExpInfoFields, 'Field must be a member of ExperimentRecordInfo.FIELDS'
        if field == ExpInfoFields.STATUS:
            assert value in ExpStatusOptions, 'Status value must be in: {}'.format(ExpStatusOptions)
        with self.persistent_obj as pod:
            pod[field] = value
        if self._text_path is not None:
            with open(self._text_path, 'w') as f:
                f.write(self.get_text())

    def add_note(self, note):
        if not self.has_field(ExpInfoFields.NOTES):
            self.set_field(ExpInfoFields.NOTES, [note])
        else:
            self.set_field(ExpInfoFields.NOTES, self.get_field(ExpInfoFields.NOTES) + [note])

    def get_text(self):
        if ExpInfoFields.VERSION not in self.persistent_obj:  # Old version... we must adapt
            return '\n'.join(
                '{}: {}'.format(key, self.get_field_text(key)) for key, value in self.persistent_obj.iteritems())
        else:
            return '\n'.join(
                '{}: {}'.format(key.value, self.get_field_text(key)) for key, value in self.persistent_obj.iteritems())

    def get_field_text(self, field, replacement_if_none=''):
        assert field in ExpInfoFields, 'Field must be a member of ExperimentRecordInfo.FIELDS'
        if not self.has_field(field):
            return replacement_if_none
        elif field is ExpInfoFields.STATUS:
            return self.get_field(field).value
        elif field is ExpInfoFields.ARGS:
            return ['{}={}'.format(k, v) for k, v in self.get_field(field)]
        else:
            return str(self.get_field(field))


class NoSavedResultError(Exception):
    def __init__(self, experiment_record_id):
        Exception.__init__(self, "Experiment Record {} has no saved result.".format(experiment_record_id))


class ExperimentRecord(object):

    ERROR_FILE_NAME = 'errortrace.txt'

    def __init__(self, experiment_directory):
        self._experiment_directory = experiment_directory
        self._info = ExperimentRecordInfo(os.path.join(experiment_directory, 'info.pkl'))

    @property
    def info(self):
        return self._info

    def show_figures(self):
        from artemis.plotting.saving_plots import show_saved_figure
        for loc in self.get_figure_locs():
            show_saved_figure(loc)

    def get_log(self):
        log_file_path = os.path.join(self._experiment_directory, 'output.txt')
        assert os.path.exists(log_file_path), 'No output file found.  Maybe "%s" is not an experiment directory?' % (
        self._experiment_directory,)
        with open(log_file_path) as f:
            text = f.read()
        return text

    def list_files(self, full_path=False):
        """
        List files in experiment directory, relative to root.
        :param full_path: If true, list file with the full local path
        :return: A list of strings indicating the file paths.
        """
        paths = [os.path.join(root, filename) for root, _, files in os.walk(self._experiment_directory) for filename in
                 files]
        if not full_path:
            dir_length = len(self._experiment_directory)
            paths = [f[dir_length + 1:] for f in paths]
        return paths

    def open_file(self, filename, *args, **kwargs):
        """
        Open a file within the experiment record folder.
        Example Usage:

            with record.open_file('filename.txt') as f:
                txt = f.read()

        :param filename: Path within experiment directory (it can include subdirectories)
        :param args, kwargs: Forwarded to python's "open" function
        :return: A file object
        """
        full_path = os.path.join(self._experiment_directory, filename)
        make_file_dir(full_path)  # Make the path if it does not yet exist
        return open(full_path, *args, **kwargs)

    def get_figure_locs(self, include_directory=True):
        locs = [f for f in os.listdir(self._experiment_directory) if f.startswith('fig-')]
        if include_directory:
            return [os.path.join(self._experiment_directory, f) for f in locs]
        else:
            return locs

    def show(self):
        print '{header} Showing Experiment {header}\n{info}\n{subborder}Logs {subborder}\n{log}\n{border}'.format(
            header="=" * 20, border="=" * 50, info=self.info.get_text(), subborder='-' * 20, log=self.get_log())
        self.show_figures()

    def get_info_text(self):
        return self.info.get_text()

    def has_result(self):
        return os.path.exists(os.path.join(self._experiment_directory, 'result.pkl'))

    def get_result(self):
        result_loc = os.path.join(self._experiment_directory, 'result.pkl')
        if os.path.exists(result_loc):
            with open(result_loc) as f:
                result = pickle.load(f)
            return result
        else:
            raise NoSavedResultError(self.get_identifier())

    def save_result(self, result):
        file_path = get_local_experiment_path(os.path.join(self._experiment_directory, 'result.pkl'))
        make_file_dir(file_path)
        with open(file_path, 'w') as f:
            pickle.dump(result, f, protocol=2)
            print 'Saving Result for Experiment "%s"' % (self.get_identifier(),)

    def get_identifier(self):
        root, identifier = os.path.split(self._experiment_directory)
        return identifier

    def get_one_liner(self):
        """
        :return: A one line description of the experiment resutls
        """
        if not self.has_result():
            return '<Experiment did not finish>'
        else:
            result = self.get_result()
            if result is None:
                return '<Experiment Returned no result>'
            else:
                try:
                    exp = self.get_experiment()
                except ExperimentNotFoundError:
                    return '<Experiment {} was not found>'.format(record_id_to_experiment_id(self.get_identifier()))
                except Exception as err:
                    return '<Error loading experiment>'
                one_liner = exp.get_one_liner(result)
                if one_liner is None:
                    return '<One-liner function not defined>'
                else:
                    return one_liner

    def get_experiment(self):
        return load_experiment(record_id_to_experiment_id(self.get_identifier()))

    def get_name(self):
        return self.get_identifier()[27:]  # NOTE: THIS WILL HAVE TO CHANGE IF WE USE A DIFFERENT DATA FORMAT!

    def get_dir(self):
        return self._experiment_directory

    def delete(self):
        shutil.rmtree(self._experiment_directory)

    @classmethod
    def from_identifier(cls, record_id):
        path = os.path.join(get_local_path(os.path.join('experiments', record_id)))
        return ExperimentRecord(path)

    def get_invalid_arg_note(self):
        """
        Return a string identifying ig the arguments for this experiment are still valid.
        :return:
        """
        experiment_id = record_id_to_experiment_id(self.get_identifier())
        if is_experiment_loadable(experiment_id):
            last_run_args = dict(self.info.get_field(ExpInfoFields.ARGS))
            current_args = dict(load_experiment(record_id_to_experiment_id(self.get_identifier())).get_args())
            validity = self.is_valid(last_run_args=last_run_args, current_args=current_args)
            if validity is False:
                last_arg_str, this_arg_str = [
                    ['{}:{}'.format(k, v) for k, v in argdict.iteritems()] if isinstance(argdict, dict) else
                    ['{}:{}'.format(k, v) for k, v in argdict]
                    for argdict in (last_run_args, current_args)]
                common, (old_args, new_args) = separate_common_items([last_arg_str, this_arg_str])
                notes = "No: Args changed!: {{{}}}->{{{}}}".format(','.join(old_args), ','.join(new_args))
            elif validity is None:
                notes = "Cannot Determine."
            else:
                notes = "Yes"
        else:
            notes = "<Experiment Not Currently Imported>"
        return notes

    def is_valid(self, last_run_args=None, current_args=None):
        """
        :return: True if the experiment arguments have not changed
            False if they have changed
            None if it cannot be determined because arguments are not hashable objects.
        """
        if last_run_args is None:
            last_run_args = dict(self.info.get_field(ExpInfoFields.ARGS))
        if current_args is None:
            current_args = dict(load_experiment(record_id_to_experiment_id(self.get_identifier())).get_args())
        try:
            return compute_fixed_hash(last_run_args, try_objects=True) == compute_fixed_hash(current_args, try_objects=True)
        except NotImplementedError:  # Happens when we have unhashable arguments
            return None

    def get_error_trace(self):
        """
        Get the error trace, or return None if there is no error trace.
        :return:
        """
        file_path = os.path.join(self._experiment_directory, self.ERROR_FILE_NAME)
        if os.path.exists(file_path):
            with open(file_path) as f:
                return f.read()
        else:
            return None

    def write_error_trace(self, print_too = True):
        file_path = os.path.join(self._experiment_directory, self.ERROR_FILE_NAME)
        assert not os.path.exists(file_path), 'Error trace has already been created in this experiment... Something fishy going on here.'
        with open(file_path, 'w') as f:
            error_text = traceback.format_exc()
            f.write(error_text)
        if print_too:
            print error_text

_CURRENT_EXPERIMENT_RECORD = None


@contextmanager
def record_experiment(identifier='%T-%N', name='unnamed', print_to_console=True, show_figs=None,
                      save_figs=True, saved_figure_ext='.pdf', use_temp_dir=False, date=None):
    """
    :param identifier: The string that uniquely identifies this experiment record.  Convention is that it should be in
        the format
    :param name: Base-name of the experiment
    :param print_to_console: If True, print statements still go to console - if False, they're just rerouted to file.
    :param show_figs: Show figures when the experiment produces them.  Can be:
        'hang': Show and hang
        'draw': Show but keep on going
        False: Don't show figures
    """
    # Note: matplotlib imports are internal in order to avoid trouble for people who may import this module without having
    # a working matplotlib (which can occasionally be tricky to install).
    if date is None:
        date = datetime.now()
    identifier = format_filename(file_string=identifier, base_name=name, current_time=date)

    if show_figs is None:
        show_figs = 'draw' if is_test_mode() else 'hang'

    assert show_figs in ('hang', 'draw', False)

    if use_temp_dir:
        experiment_directory = tempfile.mkdtemp()
        atexit.register(lambda: shutil.rmtree(experiment_directory))
    else:
        experiment_directory = get_local_path('experiments/{identifier}'.format(identifier=identifier))

    make_dir(experiment_directory)
    from artemis.plotting.manage_plotting import WhatToDoOnShow
    global _CURRENT_EXPERIMENT_RECORD  # Register
    _CURRENT_EXPERIMENT_RECORD = ExperimentRecord(experiment_directory)
    capture_context = CaptureStdOut(log_file_path=os.path.join(experiment_directory, 'output.txt'),
                                    print_to_console=print_to_console)
    show_context = WhatToDoOnShow(show_figs)
    if save_figs:
        from artemis.plotting.saving_plots import SaveFiguresOnShow
        save_figs_context = SaveFiguresOnShow(path=os.path.join(experiment_directory, 'fig-%T-%L' + saved_figure_ext))
        with capture_context, show_context, save_figs_context:
            yield _CURRENT_EXPERIMENT_RECORD
    else:
        with capture_context, show_context:
            yield _CURRENT_EXPERIMENT_RECORD
    _CURRENT_EXPERIMENT_RECORD = None  # Deregister


def get_current_experiment_record():
    if _CURRENT_EXPERIMENT_RECORD is None:
        raise Exception("No experiment is currently running!")
    return _CURRENT_EXPERIMENT_RECORD


def get_current_experiment_id():
    """
    :return: A string identifying the current experiment
    """
    return get_current_experiment_record().get_identifier()


def get_current_experiment_name():
    """
    :return: A string containing the name of the current experiment
    """
    return get_current_experiment_record().get_name()


def get_current_experiment_dir():
    """
    The directory in which the results of the current experiment are recorded.
    """
    return get_current_experiment_record().get_dir()


def open_in_experiment_dir(filename, *args, **kwargs):
    """
    Open a file in the given experiment directory.  Usage:

    with open_in_experiment_dir('myfile.txt', 'w') as f:
        f.write('blahblahblah')

    :param filename: The name of the file, relative to your experiment directory,
    :param args,kwargs: See python built-in "open" function
    :yield: The file object
    """
    return get_current_experiment_record().open_file(filename, *args, **kwargs)


def run_experiment(name, exp_dict=GLOBAL_EXPERIMENT_LIBRARY, **experiment_record_kwargs):
    """
    Run an experiment and save the results.  Return a string which uniquely identifies the experiment.
    You can run the experiment agin later by calling show_experiment(location_string):

    :param name: The name for the experiment (must reference something in exp_dict)
    :param exp_dict: A dict<str:func> where funcs is a function with no arguments that run the experiment.
    :param experiment_record_kwargs: Passed to ExperimentRecord.

    :return: A location_string, uniquely identifying the experiment.
    """
    experiment = exp_dict[name]
    return experiment.run(**experiment_record_kwargs)


def record_id_to_experiment_id(record_id):
    return record_id[27:]


def record_id_to_timestamp(record_id):
    return record_id[:26]


def run_experiment_ignoring_errors(name, **kwargs):
    try:
        run_experiment(name, **kwargs)
    except Exception as err:
        traceback.print_exc()


def _get_matching_template_from_experiment_name(experiment_name, template='%T-%N'):
    # Potentially obselete, though will keep around in case it gets useful one day.
    named_template = template.replace('%N', re.escape(experiment_name))
    expr = named_template.replace('%T', '\d\d\d\d\.\d\d\.\d\d\T\d\d\.\d\d\.\d\d\.\d\d\d\d\d\d')
    expr = '^' + expr + '$'
    return expr


def delete_experiment_with_id(experiment_identifier):
    if experiment_exists(experiment_identifier):
        load_experiment_record(experiment_identifier).delete()


def get_local_experiment_path(identifier):
    return os.path.join(get_local_path('experiments'), identifier)


def experiment_exists(identifier):
    local_path = get_local_experiment_path(identifier)
    return os.path.exists(local_path)


def merge_experiment_dicts(*dicts):
    """
    Merge dictionaries of experiments, checking that names are unique.
    """
    merge_dict = OrderedDict()
    for d in dicts:
        assert not any(k in merge_dict for k in d), "Experiments %s has been defined twice." % (
        [k for k in d.keys() if k in merge_dict],)
        merge_dict.update(d)
    return merge_dict


def filter_experiment_ids(ids, expr=None, names=None):
    if expr is not None:
        ids = [e for e in ids if expr in e]
    if names is not None:
        ids = [eid for eid in ids if record_id_to_experiment_id(eid) in names]
    return ids


def get_all_record_ids(experiment_ids=None, filters=None):
    """
    :param experiment_ids: A list of experiment names
    :param filters: A list or regular expressions for matching experiments.
    :return: A list of experiment identifiers.
    """
    expdir = get_local_path('experiments')
    ids = [e for e in os.listdir(expdir) if os.path.isdir(os.path.join(expdir, e))]
    ids = filter_experiment_ids(ids=ids, names=experiment_ids)
    if filters is not None:
        for expr in filters:
            ids = filter_experiment_ids(ids=ids, expr=expr)
    ids = sorted(ids)
    return ids


def experiment_id_to_record_ids(experiment_identifier, filter_status = None):
    """
    :param experiment_identifier: The name of the experiment
    :param filter_status: An ExpStatusOptions enum, indicating that you only want experiments with this status.
        e.g. ExpStatusOptions.FINISHED
    :return: A list of records for this experiment, temporal order
    """
    matching_records = get_all_record_ids(experiment_ids=[experiment_identifier])
    if filter_status is not None:
        assert filter_status in ExpStatusOptions, 'filter status must be one of the ExpStatusOptions'
        matching_records = [record for record in  matching_records if load_experiment_record(record).info.get_field(ExpInfoFields.STATUS) is filter_status]
    return sorted(matching_records)


def experiment_id_to_latest_record_id(experiment_identifier, filter_status = None, actual_index=-1):
    """
    Show results of the latest experiment matching the given template.
    :param name: The experiment name
    :param template: The template which turns a name into an experiment identifier
    :return: A string identifying the latest matching experiment, or None, if not found.
    """

    all_records = experiment_id_to_record_ids(experiment_identifier, filter_status=filter_status)
    return all_records[actual_index] if len(all_records)>0 else None


def experiment_id_to_latest_result(experiment_id):
    return load_latest_experiment_record(experiment_id).get_result()


def load_latest_experiment_record(experiment_name, filter_status=None, actual_index=-1):
    experiment_record_identifier = experiment_id_to_latest_record_id(experiment_name, filter_status=filter_status, actual_index=actual_index)
    return None if experiment_record_identifier is None else load_experiment_record(experiment_record_identifier)


def has_experiment_record(experiment_identifier):
    return os.path.exists(get_local_experiment_path(identifier=experiment_identifier))


def load_experiment_record(experiment_identifier):
    """
    Load an ExperimentRecord based on the identifier
    :param experiment_identifier: A string identifying the experiment
    :return: An ExperimentRecord object
    """
    full_path = get_local_experiment_path(identifier=experiment_identifier)
    return ExperimentRecord(full_path)


def _register_experiment(experiment):
    GLOBAL_EXPERIMENT_LIBRARY[experiment.name] = experiment


def clear_experiment_records(ids=None):
    """
    Delete all experiments with ids in the list, or all experiments if ids is None.
    :param ids: A list of experiment ids, or None to remove all.
    """
    # Credit: http://stackoverflow.com/questions/185936/delete-folder-contents-in-python
    folder = get_local_path('experiments')

    if ids is None:
        ids = os.listdir(folder)

    for exp_id in ids:
        exp_path = os.path.join(folder, exp_id)
        try:
            if os.path.isfile(exp_path):
                os.unlink(exp_path)
            elif os.path.isdir(exp_path):
                shutil.rmtree(exp_path)
        except Exception as e:
            print(e)


def get_experiment_info(name):
    experiment = load_experiment(name)
    return str(experiment)


class ExperimentNotFoundError(Exception):
    def __init__(self, experiment_id):
        Exception.__init__(self,
                           'Experiment "{}" could not be loaded, either because it has not been imported, or its definition was removed.'.format(
                               experiment_id))


def load_experiment(experiment_id):
    try:
        return GLOBAL_EXPERIMENT_LIBRARY[experiment_id]
    except KeyError:
        raise ExperimentNotFoundError(experiment_id)


def is_experiment_loadable(experiment_id):
    return experiment_id in GLOBAL_EXPERIMENT_LIBRARY


def _kwargs_to_experiment_name(kwargs):
    return ','.join('{}={}'.format(argname, kwargs[argname]) for argname in sorted(kwargs.keys()))


keep_record_by_default = None


@contextmanager
def experiment_testing_context():
    """
    Use this context when testing the experiment/experiment_record infrastructure.
    Should only really be used in test_experiment_record.py
    """
    ids = get_all_record_ids()
    global keep_record_by_default
    old_val = keep_record_by_default
    keep_record_by_default = True
    yield
    keep_record_by_default = old_val

    def clean_on_close():
        new_ids = set(get_all_record_ids()).difference(ids)
        clear_experiment_records(list(new_ids))

    atexit.register(
        clean_on_close)  # We register this on exit to avoid race conditions with system commands when we open figures externally


def save_figure_in_experiment(name, fig=None, default_ext='.pdf'):
    '''
    Saves the given figure in the experiment directory. If no figure is passed, plt.gcf() is saved instead.
    :param name: The name of the figure to be saved
    :param fig: The figure to be saved, can be None
    :param default_ext: See artemis.plotting.saving_plots.save_figure() for information
    :return: The path to the figure
    '''
    import matplotlib.pyplot as plt
    from artemis.plotting.saving_plots import save_figure
    if fig is None:
        fig = plt.gcf()
    save_path = os.path.join(get_current_experiment_dir(), name)
    save_figure(fig, path=save_path, default_ext=default_ext)
    return save_path


class Experiment(object):
    """
    An experiment.  In general you should not use this class directly.  Use the experiment_function decorator, and
    create variants using decorated_function.add_variant()
    """

    def __init__(self, function=None, display_function=None, comparison_function=None, one_liner_results=None,
                 name=None, is_root=False):
        """
        :param function: The function defining the experiment
        :param display_function: A function that can be called to display the results returned by function.
            This can be useful if you just want to re-run the display for already-computed and saved results.
            To do this, go experiment.save_last()
        :param conclusion: <Deprecated> will be removed in future
        :param name: Nmae of this experiment.
        """
        self.name = name
        self.function = function
        self.display_function = display_function
        self.one_liner_results = one_liner_results
        self.comparison_function = comparison_function
        self.variants = OrderedDict()
        self._notes = []
        self.is_root = is_root
        if not is_root:
            _register_experiment(self)

    def __call__(self, *args, **kwargs):
        """ Run the function as normal, without recording or anything.  You can also modify with arguments. """
        return self.function(*args, **kwargs)

    def __str__(self):
        return 'Experiment {}'.format(self.name)

    def get_args(self):
        """
        :param to_root: If True, find all args of this experiment down to the root experiment.
            If False, just return the args that differentiate this variant from its parent.
        :return: A dictionary of arguments to the experiment
        """
        return infer_derived_arg_values(self.function)

    def get_root_function(self):
        return get_partial_chain(self.function)[0]

    def run(self, print_to_console=True, show_figs=None, test_mode=None, keep_record=None, raise_exceptions=True,
            display_results=True, **experiment_record_kwargs):
        """
        Run the experiment, and return the ExperimentRecord that is generated.

        :param print_to_console: Print to console (as well as logging to file)
        :param show_figs: Show figures (as well as saving to file)
        :param test_mode: Run in "test_mode".  This sets the global "test_mode" flag when running the experiment.  This
            flag can be used to, for example, shorten a training session to verify that the code runs.  Can be:
                True: Run in test mode
                False: Don't run in test mode:
                None: Keep the current state of the global "is_test_mode()" flag.
        :param keep_record: Keep the folder that results are saved into.
                True: Results are saved into a folder
                False: Results folder is deleted at the end.
                None: If "test_mode" is true, then delete results at end, otherwise save them.
        :param raise_exceptions: True to raise any exception that occurs when running the experiment.  False to catch it,
            print the error, and move on.
        :param experiment_record_kwargs: Passed to the "record_experiment" context.
        :return: The ExperimentRecord object, if keep_record is true, otherwise None
        """
        if test_mode is None:
            test_mode = is_test_mode()
        if keep_record is None:
            keep_record = keep_record_by_default if keep_record_by_default is not None else not test_mode

        old_test_mode = is_test_mode()
        set_test_mode(test_mode)
        ARTEMIS_LOGGER.info('{border} {mode} Experiment: {name} {border}'.format(border='=' * 10,
                                                                                 mode="Testing" if test_mode else "Running",
                                                                                 name=self.name))
        EIF = ExpInfoFields
        date = datetime.now()
        with record_experiment(name=self.name, print_to_console=print_to_console, show_figs=show_figs,
                use_temp_dir=not keep_record, date=date, **experiment_record_kwargs) as exp_rec:
            start_time = time.time()
            try:

                exp_rec.info.set_field(ExpInfoFields.NAME, self.name)
                exp_rec.info.set_field(ExpInfoFields.ID, exp_rec.get_identifier())
                exp_rec.info.set_field(ExpInfoFields.DIR, exp_rec.get_dir())
                exp_rec.info.set_field(EIF.ARGS, self.get_args().items())
                root_function = self.get_root_function()
                exp_rec.info.set_field(EIF.FUNCTION, root_function.__name__)
                exp_rec.info.set_field(EIF.TIMESTAMP, str(date))
                module = inspect.getmodule(root_function)
                exp_rec.info.set_field(EIF.MODULE, module.__name__)
                exp_rec.info.set_field(EIF.FILE, module.__file__ if hasattr(module, '__file__') else '<unknown>')
                exp_rec.info.set_field(EIF.STATUS, ExpStatusOptions.STARTED)
                exp_rec.info.set_field(EIF.USER, getuser())
                exp_rec.info.set_field(EIF.MAC, ':'.join(("%012X" % getnode())[i:i+2] for i in range(0, 12, 2)))
                results = self.function()
                exp_rec.info.set_field(EIF.STATUS, ExpStatusOptions.FINISHED)
            except KeyboardInterrupt:
                exp_rec.info.set_field(EIF.STATUS, ExpStatusOptions.STOPPED)
                exp_rec.write_error_trace(print_too=False)
                raise
            except Exception:
                exp_rec.info.set_field(EIF.STATUS, ExpStatusOptions.ERROR)
                exp_rec.write_error_trace(print_too=not raise_exceptions)
                if raise_exceptions:
                    raise
                else:
                    return exp_rec
            finally:
                exp_rec.info.set_field(EIF.RUNTIME, time.time() - start_time)
                fig_locs = exp_rec.get_figure_locs(include_directory=False)
                exp_rec.info.set_field(EIF.N_FIGS, len(fig_locs))
                exp_rec.info.set_field(EIF.FIGS, fig_locs)

        exp_rec.save_result(results)
        for n in self._notes:
            exp_rec.info.add_note(n)
        if display_results and self.display_function is not None:
            self.display_function(results)
        ARTEMIS_LOGGER.info('{border} Done {mode} Experiment: {name} {border}'.format(border='=' * 10, mode="Testing" if test_mode else "Running", name=self.name))
        set_test_mode(old_test_mode)
        return exp_rec

    def _create_experiment_variant(self, args, kwargs, is_root):
        assert len(args) in (0, 1), "When creating an experiment variant, you can either provide one unnamed argument (the experiment name), or zero, in which case the experiment is named after the named argumeents.  See add_variant docstring"
        name = args[0] if len(args) == 1 else _kwargs_to_experiment_name(kwargs)
        assert name not in self.variants, 'Variant "%s" already exists.' % (name,)
        ex = Experiment(
            name=self.name + '.' + name,
            function=partial(self.function, **kwargs),
            display_function=self.display_function,
            comparison_function=self.comparison_function,
            one_liner_results=self.one_liner_results,
            is_root=is_root
        )
        self.variants[name] = ex
        return ex

    def add_variant(self, *args, **kwargs):
        """
        Add a variant to this experiment, and register it on the list of experiments.
        There are two ways you can do this:

            # Name the experiment explicitely, then list the named arguments
            my_experiment_function.add_variant('big_a', a=10000)
            assert my_experiment_function.get_name()=='my_experiment_function.big_a'

            # Allow the experiment to be named automatically, and just list the named arguments
            my_experiment_function.add_variant(a=10000)
            assert my_experiment_function.get_name()=='my_experiment_function.a==10000'

        :return: The experiment.
        """
        return self._create_experiment_variant(args, kwargs, is_root=False)

    def add_root_variant(self, *args, **kwargs):
        """
        Add a variant to this experiment, but do NOT register it on the list of experiments.
        There are two ways you can do this:

            # Name the experiment explicitely, then list the named arguments
            my_experiment_function.add_root_variant('big_a', a=10000)
            assert my_experiment_function.get_name()=='my_experiment_function.big_a'

            # Allow the experiment to be named automatically, and just list the named arguments
            my_experiment_function.add_root_variant(a=10000)
            assert my_experiment_function.get_name()=='my_experiment_function.a==10000'

        :return: The experiment.
        """
        return self._create_experiment_variant(args, kwargs, is_root=True)

    def add_note(self, note):
        """
        :param note:
        :return:
        """
        self._notes.append(str(note))
        return self

    def get_variant(self, *args, **kwargs):
        """
        Get a variant on this experiment.
        :param name: A the name of the variant
        :param path: Optionally, a list of names of subvariants (to call up a nested experiment)
        :return:
        """
        if len(args)==0:
            name = _kwargs_to_experiment_name(kwargs)
        else:
            assert len(args)==1, 'You can only provide 1 unnamed argument to get_variant: the variant name.'
            name, = args
            assert len(kwargs)==0, 'If you provide a variant name ({}), there is no need to specify the keyword arguments. ({})'.format(name, kwargs)
        assert name in self.variants, "No variant '{}' exists.  Existing variants: {}".format(name, self.variants.keys())
        return self.variants[name]

    def get_unnamed_variant(self, **kwargs):
        return self.get_variant(_kwargs_to_experiment_name(kwargs))

    def display_last(self, result='___FINDLATEST', err_if_none=True):
        if isinstance(result, basestring) and result == '___FINDLATEST':
            result = experiment_id_to_latest_result(self.name)
        if err_if_none:
            assert result is not None, "No result was computed for the last run of '%s'" % (self.name,)
        if result is None:
            if err_if_none:
                raise Exception("No result was computed for the last run of '%s'" % (self.name,))
            else:
                print "<No result saved from last run>"
        elif self.display_function is None:
            print result
        else:
            self.display_function(result)

    def get_one_liner(self, results):
        if self.one_liner_results is not None:
            return self.one_liner_results(results)
        elif isinstance(results, np.ndarray):
            short = 'ndarray<shape={},dtype={}>'.format(results.shape, results.dtype).replace(' ', '')
            if results.size<4:
                return short+str(results).replace('\n', ';')
            else:
                return short
        else:
            return str(results).replace('\n', ';')

    def display_or_run(self):
        """
        Display the last results, or, if the experiment has not been run yet, run it and then display the results.
        A word of caution: This function does NOT check that the parameters of the last experiment are the same as the
        current parameters.

        :return:
        """
        if experiment_id_to_latest_record_id(self.name) is None:
            self.run()
        else:
            result = experiment_id_to_latest_result(self.name)
            if result is not None and self.display_function is not None:
                self.display_last()
            else:
                self.run()

    def get_all_variants(self, include_roots=False, include_self=False):
        """
        Return a list of variants of this experiment
        :param include_roots: Include "root" experiments
        :param include_self: Include this experiment (unless include_roots is false and this this experiment is a root)
        :return: A list of experiments.
        """
        variants = []
        if include_self and (not self.is_root or include_roots):
            variants.append(self)
        for name, v in self.variants.iteritems():
            variants += v.get_all_variants(include_roots=include_roots, include_self=True)
        return variants

    def run_all(self, include_roots = False):
        """
        Run this experiment (if not a root-experiment) and all variants (if not roots).
        """
        experiments = self.get_all_variants(include_self=True, include_roots=include_roots)
        for ex in experiments:
            ex.run()

    def run_all_multiprocess(self):
        experiments = self.get_all_variants()
        p = multiprocessing.Pool(processes=multiprocessing.cpu_count())
        p.map(run_experiment, [ex.name for ex in experiments])

    def test(self, **kwargs):
        self.run(test_mode=True, **kwargs)

    def test_all(self, **kwargs):
        self.run_all(test_mode=True, **kwargs)

    def compare_results(self, experiment_ids, error_if_no_result = True):
        if self.comparison_function is None:
            print 'Cannot compare results, because you have not specified any comparison function for this experiment.  Use @ExperimentFunction(comparison_function = my_func)'
            return
        results = load_lastest_experiment_results(experiment_ids, error_if_no_result=error_if_no_result)
        self.comparison_function(results)

    def get_all_ids(self):
        """
        Get all identifiers of this experiment that have been run.
        :return:
        """
        return get_all_record_ids(experiment_ids=[self.name])

    def clear_records(self, include_children=False):
        """
        Delete all records from this experiment
        """
        clear_experiment_records(ids=self.get_all_ids())
        if include_children:
            children = self.get_all_variants(include_roots=False, include_self=True)
            for child in children:
                child.clear_records()

    def get_name(self):
        return self.name


def load_lastest_experiment_results(experiment_ids, error_if_no_result = True):
    results = OrderedDict()
    for eid in experiment_ids:
        record = load_latest_experiment_record(eid, filter_status=ExpStatusOptions.FINISHED)
        if record is None:
            if error_if_no_result:
                raise Exception("Experiment {} had no result.  Run this experiment to completion before trying to compare its results.".format(eid))
            else:
                ARTEMIS_LOGGER.warn('Experiment {} had no records.  Not including this in results'.format(eid))
        else:
            results[eid] = record.get_result()
    if len(results)==0:
        ARTEMIS_LOGGER.warn('None of your experiments had any results.  Your comparison function will probably show no meaningful result.')
    return results


def make_record_comparison_table(record_ids, args_to_show=None, results_extractor = None, print_table = False):
    """
    Make a table comparing the arguments and results of different experiment records.  You can use the output
    of this function with the tabulate package to make a nice readable table.

    :param record_ids: A list of record ids whose results to compare
    :param args_to_show: A list of arguments to show.  If none, it will just show all arguments
        that differ between experiments.
    :param results_extractor: A dict<str->callable> where the callables take the result of the
        experiment as an argument and return an entry in the table.  For example:
    :param print_table: Optionally, import tabulate and print the table here and now.
    :return: headers, rows
        headers is a list of of headers for the top of the table
        rows is a list of lists filling in the information.

    example usage:

        headers, rows = make_record_comparison_table(
            record_ids = [experiment_id_to_latest_record_id(eid) for eid in [
                'demo_fast_weight_mlp.multilayer_baseline.1epoch.version=mlp',
                'demo_fast_weight_mlp.multilayer_baseline.1epoch.full-gd.n_steps=1',
                'demo_fast_weight_mlp.multilayer_baseline.1epoch.full-gd.n_steps=20',
                ]],
            results_extractor={
                'Test': lambda result: result.get_best('test').score.get_score('test'),
                'Train': lambda result: result.get_best('test').score.get_score('train'),
                }
             )
        import tabulate
        print tabulate.tabulate(rows, headers=headers, tablefmt=tablefmt)
    """

    records = [load_experiment_record(rid) for rid in record_ids]
    args = [rec.info.get_field(ExpInfoFields.ARGS) for rec in records]
    if args_to_show is None:
        common, separate = separate_common_items(args)
        args_to_show = [k for k, v in separate[0]]

    if results_extractor is None:
        results_extractor = {'Result': str}
    else:
        assert isinstance(results_extractor, dict)

    headers = args_to_show + results_extractor.keys()

    rows = []
    for record, record_args in izip_equal(records, args):
        arg_dict = dict(record_args)
        args_vals = [arg_dict[k] for k in args_to_show]
        results = record.get_result()
        rows.append(args_vals+[f(results) for f in results_extractor.values()])

    if print_table:
        import tabulate
        print tabulate.tabulate(rows, headers=headers, tablefmt='simple')


    return headers, rows


def clear_all_experiments():
    GLOBAL_EXPERIMENT_LIBRARY.clear()


@contextmanager
def capture_created_experiments():
    """
    A convenient way to cross-breed experiments.  If you define experiments in this block, you can capture them for
    later use (for instance by modifying them)
    :return:
    """
    current_len = len(GLOBAL_EXPERIMENT_LIBRARY)
    new_experiments = []
    yield new_experiments
    for ex in GLOBAL_EXPERIMENT_LIBRARY.values()[current_len:]:
        new_experiments.append(ex)



import pytest

from artemis.general.mymath import softmax, cummean, cumvar, sigm, expected_sigm_of_norm, mode, cummode, normalize, is_parallel, \
    align_curves, angle_between, fixed_diff, decaying_cumsum, geosum
import numpy as np
__author__ = 'peter'


def test_softmax():
    x = np.random.randn(3, 4, 5)

    s = softmax(x, axis=1)
    assert s.shape==(3, 4, 5) and (s>0).all() and (s<1).all() and np.allclose(np.sum(s, axis=1), 1)


def test_cummean():

    arr = np.random.randn(3, 4)
    cum_arr = cummean(arr, axis = 1)
    assert np.allclose(cum_arr[:, 0], arr[:, 0])
    assert np.allclose(cum_arr[:, 1], np.mean(arr[:, :2], axis = 1))
    assert np.allclose(cum_arr[:, 2], np.mean(arr[:, :3], axis = 1))


def test_cumvar():

    arr = np.random.randn(3, 4)
    cum_arr = cumvar(arr, axis = 1, sample = False)
    assert np.allclose(cum_arr[:, 0], 0)
    assert np.allclose(cum_arr[:, 1], np.var(arr[:, :2], axis = 1))
    assert np.allclose(cum_arr[:, 2], np.var(arr[:, :3], axis = 1))


def test_exp_sig_of_norm():

    mean = 1
    std = 0.8
    n_points = 1000
    seed = 1234

    inputs = np.random.RandomState(seed).normal(mean, std, size = n_points)
    vals = sigm(inputs)
    sample_mean = np.mean(vals)

    for method in ('maclauren-2', 'maclauren-3', 'probit'):
        approx_true_mean = expected_sigm_of_norm(mean, std, method = method)
        approx_sample_mean = expected_sigm_of_norm(np.mean(inputs), np.std(inputs), method = method)
        true_error = np.abs(approx_true_mean-sample_mean)/sample_mean
        sample_error = np.abs(approx_sample_mean-sample_mean)/sample_mean
        print 'Error for %s: %.4f True, %.4f Sample.' % (method, true_error, sample_error)
        assert true_error < 0.02, 'Method %s did pretty bad' % (method, )


def test_mode():

    arr = np.random.RandomState(0).randint(low=0, high=2, size=(3, 5, 7))

    m0 = mode(arr, axis = 0)
    assert m0.shape == (5, 7)
    assert np.all(np.sum(m0[None, :, :] == arr, axis = 0) > np.sum(m0[None, :, :] != arr, axis = 0))

    m1 = mode(arr, axis = 1)
    assert m1.shape == (3, 7)
    assert np.all(np.sum(m1[:, None, :] == arr, axis = 1) > np.sum(m1[:, None, :] != arr, axis = 1))

    m2 = mode(arr, axis = 2)
    assert m2.shape == (3, 5)


@pytest.mark.skipif(True, reason='Requires scipy weave, which does not install reliably.')
def test_cummode():

    arr = np.random.RandomState(0).randint(low=0, high=3, size=(5, 7))

    m = cummode(arr, axis = 1)  # (n_samples, n_events)
    assert m.shape == arr.shape

    uniques = np.unique(arr)

    for j in xrange(arr.shape[1]):
        n_elements_of_mode_class = np.sum(arr[:, :j+1] == m[:, j][:, None], axis = 1)  # (n_samples, )
        for k, u in enumerate(uniques):
            n_elements_of_this_class = np.sum(arr[:, :j+1] == u, axis = 1)  # (n_samples, )
            assert np.all(n_elements_of_mode_class >= n_elements_of_this_class)


@pytest.mark.skipif(True, reason='Requires scipy weave, which does not install reliably.')
def test_cummode_weighted():

    arr = np.random.RandomState(0).randint(low=0, high=3, size=(5, 7))
    w = np.random.rand(5, 7)

    m = cummode(arr, weights=w, axis = 1)  # (n_samples, n_events)
    assert m.shape == arr.shape

    uniques = np.unique(arr)

    for j in xrange(arr.shape[1]):
        bool_ixs_of_mode_class = arr[:, :j+1] == m[:, j][:, None]  # (n_samples, j+1)
        weights_of_mode_class = np.sum(w[:, :j+1]*bool_ixs_of_mode_class, axis = 1)  # (n_samples, )

        for k, u in enumerate(uniques):
            bool_ixs_of_this_class = arr[:, :j+1] == u  # (n_samples, j+1)
            weights_of_this_class = np.sum(w[:, :j+1]*bool_ixs_of_this_class, axis = 1)  # (n_samples, )
            assert np.all(weights_of_mode_class >= weights_of_this_class)


def test_normalize():

    # L1 - positive values
    arr = np.random.rand(5, 4)
    norm_arr = normalize(arr, degree=1, axis = 1)
    assert np.allclose(norm_arr.sum(axis=1), 1)

    # L1 - with negative values
    arr = np.random.randn(5, 4)
    norm_arr = normalize(arr, degree=1, axis = 1)
    assert np.allclose(np.abs(norm_arr).sum(axis=1), 1)

    # L2
    arr = np.random.randn(5, 4)
    norm_arr = normalize(arr, degree=2, axis = 1)
    assert np.allclose(np.sqrt((norm_arr**2).sum(axis=1)), 1)

    # L1 - zero row with nan handling
    arr = np.random.rand(5, 4)
    arr[2, :] = 0
    norm_arr = normalize(arr, degree=1, axis = 1)
    assert np.all(np.isnan(norm_arr[2, :]))

    # L1 - zero row with nan handling
    arr = np.random.rand(5, 4)
    arr[2, :] = 0
    norm_arr = normalize(arr, degree=1, axis = 1, avoid_nans=True)
    assert np.allclose(np.abs(norm_arr).sum(axis=1), 1)
    assert np.allclose(norm_arr[2, :], 1./arr.shape[1])

    # L2 - zero row with nan handling
    arr = np.random.rand(5, 4)
    arr[2, :] = 0
    norm_arr = normalize(arr, degree=2, axis = 1, avoid_nans=True)
    assert np.allclose(np.sqrt((norm_arr**2).sum(axis=1)), 1)
    assert np.allclose(norm_arr[2, :], np.sqrt(1./arr.shape[1]))


def test_is_parallel():

    assert is_parallel([1, 2], [2, 4])
    assert not is_parallel([1, 2], [2, 5])
    assert is_parallel([1, 2], [2, 5], angular_tolerance=0.5)
    assert not is_parallel([1, 2], [-2, -4])


def test_align_curves():

    n_curves = 30

    n_points = [np.random.randint(20) for _ in xrange(n_curves)]

    xs = [np.sort(np.random.rand(n)) for n in n_points]
    ys = [np.random.randn(n) for n in n_points]

    new_xs, new_ys = align_curves(xs=xs, ys=ys, n_bins=25, spacing='lin')
    assert new_xs.shape == (25, )
    assert new_ys.shape == (n_curves, 25)

    new_xs, new_ys = align_curves(xs=xs, ys=ys, n_bins=25, spacing='log')
    assert new_xs.shape == (25, )
    assert new_ys.shape == (n_curves, 25)


def test_angle_between():

    a = np.array([[0, 1], [1, 1], [1, 0]])
    b = np.array([[1, 0], [1, 0], [1, 0]])
    angles = angle_between(a, b, in_degrees=True, axis=-1)
    assert np.allclose(angles, [90, 45, 0])
    assert np.allclose(angle_between([2, 1], [1, 0], in_degrees=True), np.arctan(1/2.)*180/np.pi)


def test_fixed_diff():

    a = np.random.randn(2, 3, 4)

    da = fixed_diff(a, axis=1, initial_value=0)
    assert np.array_equal(da[:, 1:, :], np.diff(a, axis=1))
    assert np.array_equal(da[:, 0, :], a[:, 0, :])

    da = fixed_diff(a, axis=1, initial_value='first')
    assert np.array_equal(da[:, 1:, :], np.diff(a, axis=1))
    assert np.array_equal(da[:, 0, :], np.zeros_like(da[:, 0, :]))

    assert np.allclose(a, np.cumsum(fixed_diff(a, axis=1), axis=1))


def test_decaying_cumsum():

    a = np.random.randn(2, 3, 4)

    ca = decaying_cumsum(a, axis=1, memory=0)
    assert np.array_equal(ca, a)

    ca = decaying_cumsum(a, axis=1, memory=.6)
    assert np.allclose(ca[:, 2, :], 0.4*(0.6**2*a[:, 0, :] + 0.6**1*a[:, 1, :] + a[:, 2, :]))


def test_geosum():
    assert geosum(0.5, t_end=4, t_start=2) == 0.5**2 + 0.5**3 + 0.5**4 == 0.4375
    assert geosum(1, t_end=4, t_start=2) == 1**2+1**3+1**4 == 3


if __name__ == '__main__':
    test_decaying_cumsum()
    test_fixed_diff()
    test_angle_between()
    test_align_curves()
    test_is_parallel()
    test_normalize()
    test_cummode_weighted()
    test_cummode()
    test_mode()
    test_exp_sig_of_norm()
    test_cumvar()
    test_cummean()
    test_softmax()
    test_geosum()



from artemis.general.numpy_helpers import get_rng
from artemis.ml.datasets.datasets import DataSet
from artemis.ml.tools.neuralnets import initialize_network_params, forward_pass


def get_synthetic_deep_data(n_samples, layer_sizes, hidden_activations='softplus', output_activation='linear', normalize = True, rng=1234):
    """
    Generate data from a randomly initialized neural network.
    :param n_samples: Number of samples to generate
    :param layer_sizes: Sizes of network layers
    :param hidden_activations: Hidden activation functions
    :param output_activation: Output activation function
    :param normalize: Normalize the output over samples (remove global mean, divide by std)
    :param rng: A random number generator or seed.
    :return: x, y
        x is an (n_samples, layer_sizes[0]) array
        y is a (n_samples, layer_sizes[-1]) array
    """
    rng = get_rng(rng)
    ws = initialize_network_params(layer_sizes=layer_sizes, mag = 'xavier-forward', include_biases=False, rng=rng)
    x = rng.randn(n_samples, layer_sizes[0])
    y = forward_pass(input_data=x, weights=ws, hidden_activations=hidden_activations, output_activation=output_activation)
    if normalize:
        y = (y - y.mean(axis=0))/y.std(axis=0)
    return x, y


def get_synthetic_deep_dataset(n_training, n_test, **kwargs):
    """
    :param n_training: Number of training samples
    :param n_test: Number of test samples
    :param kwargs: See get_synthetic_deep_data above
    :return:
    """
    x, y = get_synthetic_deep_data(**kwargs)
    return DataSet.from_xy(x, y, training_fraction=n_training/float(n_training+n_test))



from artemis.fileman.local_dir import get_local_path
import os
from artemis.fileman.persistent_print import capture_print, read_print, new_log_file, stop_capturing_print
from artemis.general.display import CaptureStdOut

__author__ = 'peter'


def test_proper_persistent_print_usage():
    """
    Here is the best, cleanest way to use persistent print
    :return:
    """
    print 'ddd'
    with CaptureStdOut() as ps:
        print 'fff'
        print 'ggg'
    print 'hhh'
    assert ps.read() == 'fff\nggg\n'


def test_proper_persistent_print_file_logging():

    log_file_path = get_local_path('tests/test_log.txt')
    with CaptureStdOut(log_file_path) as ps:
        print 'fff'
        print 'ggg'
    print 'hhh'
    assert ps.read() == 'fff\nggg\n'

    # You can verify that the log has also been written.
    log_path = ps.get_log_file_path()
    with open(log_path) as f:
        txt = f.read()
    assert txt == 'fff\nggg\n'


def test_persistent_print():

    test_log_path = capture_print()
    print 'aaa'
    print 'bbb'
    assert read_print()  == 'aaa\nbbb\n'
    stop_capturing_print()

    capture_print()
    assert read_print() == ''
    print 'ccc'
    print 'ddd'
    assert read_print()  == 'ccc\nddd\n'

    os.remove(get_local_path(test_log_path))


def test_new_log_file():
    # Just a shorthand for persistent print.

    log_file_loc = new_log_file('dump/test_file')
    print 'eee'
    print 'fff'
    stop_capturing_print()

    local_log_loc = get_local_path(log_file_loc)
    with open(local_log_loc) as f:
        text = f.read()

    assert text == 'eee\nfff\n'
    os.remove(local_log_loc)


if __name__ == '__main__':

    test_proper_persistent_print_usage()
    test_proper_persistent_print_file_logging()
    test_persistent_print()
    test_new_log_file()


import shlex
from collections import OrderedDict
from importlib import import_module
from artemis.experiments.experiment_record import GLOBAL_EXPERIMENT_LIBRARY, get_all_record_ids, clear_experiment_records, \
    ExperimentRecord, run_experiment_ignoring_errors, \
    experiment_id_to_record_ids, load_experiment_record, load_experiment, record_id_to_experiment_id, \
    record_id_to_timestamp, ExpInfoFields, ExpStatusOptions, has_experiment_record, NoSavedResultError
from artemis.general.display import IndentPrint, side_by_side
from artemis.general.should_be_builtins import separate_common_items, bad_value, detect_duplicates, \
    izip_equal, all_equal
from artemis.general.tables import build_table
from tabulate import tabulate
from functools import partial
import re

def _setup_input_memory():
    try:
        import readline  # Makes raw_input behave like interactive shell.
        # http://stackoverflow.com/questions/15416054/command-line-in-python-with-history
    except:
        pass  # readline not available


def _warn_with_prompt(message= None, prompt = 'Press Enter to continue', use_prompt=True):
    if message is not None:
        print message
    if use_prompt:
        raw_input('({}) >> '.format(prompt))

def find_experiment(*search_terms):
    """
    Find an experiment.  Invoke
    :param search_term: A term that will be used to search for an experiment.
    :return:
    """

    found_experiments = OrderedDict((name, ex) for name, ex in GLOBAL_EXPERIMENT_LIBRARY.iteritems() if all(re.search(term, name) for term in search_terms))
    if len(found_experiments)==0:
        raise Exception("None of the {} experiments matched the search: '{}'".format(len(GLOBAL_EXPERIMENT_LIBRARY), search_terms))
    elif len(found_experiments)>1:
        raise Exception("More than one experiment matched the search '{}', you need to be more specific.  Found: {}".format(search_terms, found_experiments.keys()))
    else:
        return found_experiments.values()[0]


def browse_experiments(catch_errors = False, close_after = False, just_last_record=False, raise_display_errors = False, run_args = None, keep_record = True, command=None):
    """
    Browse Experiments

    :param catch_errors: True if you want to catch any errors here
    :param close_after: Close this menu after running an experiment
    :param just_last_record: Just show the last record for the experiment
    """

    if run_args is None:
        run_args = {}
    if 'keep_record' not in run_args:
        run_args['keep_record'] = keep_record

    browser = ExperimentBrowser(catch_errors=catch_errors, close_after=close_after, just_last_record=just_last_record, raise_display_errors=raise_display_errors, run_args=run_args)
    browser.launch(command=command)


class ExperimentBrowser(object):

    QUIT = 'Quit'
    HELP_TEXT = """
This program lists the experiments that you have defined (referenced by E#) alongside the records of console output,
plots, results, referenced by (E#.R# - for example 4.1) created by running these experiments.  Command examples:

> 4                   Run experiment 4
> run 4               Run experiment 4
> run 4-6             Run experiment 4, 5, and 6
> run all             Run all experiments
> run 4-6 -s          Run experiments 4, 5, and 6 in sequence, and catch all errors.
> run 4-6 -e          Run experiments 4, 5, and 6 in sequence, and stop on errors
> run 4-6 -p          Run experiments 4, 5, and 6 in parallel processes, and catch all errors.
> call 4              Call experiment 4 (like running, but doesn't save a record)
> filter 4-6          Just show experiments 4-6
> filter has:xyz      Just show experiments with "xyz" in the name
> results 4-6         View the results experiments 4, 5, 6
> view results        View just the columns for experiment name and result
> view full           View all columns (the default view)
> show 4              Show the output from the last run of experiment 4 (if it has been run already).
> records             Browse through all experiment records.
> allruns             Toggle between showing all past runs of each experiment, and just the last one.
> compare 4.1,5.3     Compare records 4.1 and 5.3.
> select 4-6          Show the list of records belonging to experiments 4, 5, 6
> sidebyside 4.1,5.3  Display the output of record from experiments 4.1,5.3 side by side.
> delete 4-6          Delete all records from experiments 4, 5, 6.  You will be asked to confirm the deletion.
> q                   Quit.
> r                   Refresh list of experiments.

Commands 'run', 'call', 'filter' allow you to select experiments.  You can select experiments in the following ways:
    4               Select experiment #4
    4-6             Select experiments 4, 5, 6
    all             Select all experiments
    unfinished      Select all experiment for which there are no records of it being run to completion.
    invalid         Select all experiments where all records were made before arguments to the experiment have changed
    has:xyz         Select all experiments with the string "xyz" in their names
    hasnot:xyz      Select all experiments without substring "xyz" in their names

Commands 'results', 'show', 'records', 'compare', 'sidebyside', 'select', 'delete' allow you to specify a range of experiment
records.  You can specify records in the following ways:
    4.2             Select record 2 for experiment 4
    4               Select all records for experiment 4
    4-6             Select all records for experiments 4, 5, 6
    4.2-5           Select records 2, 3, 4, 5 for experiment 4
    4.3,4.4         Select records 4.3, 4.4
    all             Select all records
    old             Select all records that are not the the most recent run for that experiment
    unfinished      Select all records that have not run to completion
    invalid         Select all records for which the arguments to their experiments have changed since they were run
    errors          Select all records that ended in error
    invalid|errors  Select all records that are invalid or ended in error (the '|' can be used to "or" any of the above)
    invalid&errors  Select all records that are invalid and ended in error (the '&' can be used to "and" any of the above)
"""

    def __init__(self, catch_errors = False, close_after = True, just_last_record = False, view_mode ='full', raise_display_errors=False, run_args=None):

        self.close_after = close_after
        self.just_last_record = just_last_record
        self.catch_errors = catch_errors
        self.exp_record_dict = self.reload_record_dict()
        self.raise_display_errors = raise_display_errors
        self.view_mode = view_mode
        self._filter = None
        self.run_args = {} if run_args is None else run_args


    def reload_record_dict(self):
        d= OrderedDict((name, experiment_id_to_record_ids(name)) for name in GLOBAL_EXPERIMENT_LIBRARY.keys())
        if self.just_last_record:
            for k in d.keys():
                d[k] = [d[k][-1]] if len(d[k])>0 else []
        return d

    def launch(self, command=None):

        func_dict = {
            'run': self.run,
            'test': self.test,
            'show': self.show,
            'call': self.call,
            'select': self.select,
            'allruns': self.allruns,
            'view': self.view,
            'h': self.help,
            'results': self.results,
            'filter': self.filter,
            'explist': self.explist,
            'sidebyside': self.side_by_side,
            'compare': self.compare,
            'compare_results': self.compare_results,
            'delete': self.delete,
            'errortrace': self.errortrace,
            'q': self.quit,
            'records': self.records
            }

        while True:
            self.exp_record_dict = self.reload_record_dict()

            print "==================== Experiments ===================="
            print self.get_experiment_list_str(self.exp_record_dict, just_last_record=self.just_last_record, view_mode=self.view_mode, raise_display_errors=self.raise_display_errors, exp_filter=self._filter)
            print '-----------------------------------------------------'
            if command is None:
                user_input = raw_input('Enter command or experiment # to run (h for help) >> ').lstrip(' ').rstrip(' ')
            else:
                user_input=command
                command = None
            with IndentPrint():
                try:
                    split = user_input.split(' ')
                    if len(split)==0:
                        continue
                    cmd = split[0]
                    args = split[1:]

                    if cmd == '':
                        continue
                    elif cmd in func_dict:
                        out = func_dict[cmd](*args)
                    elif interpret_numbers(cmd) is not None:
                        if not any(x in args for x in ('-s', '-e', '-p')):
                            args = args + ['-e']
                        out = self.run(cmd, *args)
                    elif cmd == 'r':  # Refresh
                        continue
                    else:
                        response = raw_input('Unrecognised command: "{}".  Type "h" for help or Enter to continue. >'.format(cmd))
                        if response.lower()=='h':
                            self.help()
                        out = None
                    if out is self.QUIT or self.close_after:
                        break
                except Exception as name:
                    if self.catch_errors:
                        res = raw_input('%s: %s\nEnter "e" to view the stacktrace, or anything else to continue.' % (name.__class__.__name__, name.message))
                        if res == 'e':
                            raise
                    else:
                        raise

    @staticmethod
    def get_experiment_list_str(exp_record_dict, just_last_record, view_mode='full', raise_display_errors=False, exp_filter = None):

        headers = {
            'full': ['E#', 'R#', 'Name', 'Last Run' if just_last_record else 'All Runs', 'Duration', 'Status', 'Valid', 'Result'],
            'results': ['E#', 'R#', 'Name', 'Result']
            }[view_mode]

        rows = []

        def get_field(header):
            try:
                return \
                    index if header=='#' else \
                    (str(i) if j==0 else '') if header == 'E#' else \
                    j if header == 'R#' else \
                    (name if j==0 else '') if header=='Name' else \
                    experiment_record.info.get_field_text(ExpInfoFields.TIMESTAMP) if header in ('Last Run', 'All Runs') else \
                    experiment_record.info.get_field_text(ExpInfoFields.RUNTIME) if header=='Duration' else \
                    experiment_record.info.get_field_text(ExpInfoFields.STATUS) if header=='Status' else \
                    experiment_record.get_invalid_arg_note() if header=='Valid' else \
                    experiment_record.get_one_liner() if header=='Result' else \
                    '???'
            except:
                if raise_display_errors:
                    raise
                return '<Display Error>'

        exps_to_show = set(exp_record_dict.keys()) if exp_filter is None else set(select_experiments(exp_filter, exp_record_dict))

        for i, (exp_id, record_ids) in enumerate(exp_record_dict.iteritems()):
            if len(record_ids)==0:
                if exp_id in exps_to_show:
                    rows.append([str(i), '', exp_id, '<No Records>', '-', '-', '-', '-'])
            else:
                for j, record_id in enumerate(record_ids):
                    index, name = ['{}.{}'.format(i, j), exp_id] if j==0 else ['{}.{}'.format('`'*len(str(i)), j), exp_id]
                    try:
                        experiment_record = load_experiment_record(record_id)
                    except:
                        experiment_record = None
                    if exp_id in exps_to_show:
                        rows.append([get_field(h) for h in headers])
        assert all_equal([len(headers)]+[len(row) for row in rows]), 'Header length: {}, Row Lengths: \n  {}'.format(len(headers), '\n'.join([len(row) for row in rows]))
        table = tabulate(rows, headers=headers)
        if exp_filter:
            table += '\n[Filtered with "{}" to show {}/{} experiments]'.format(exp_filter, len(exps_to_show), len(exp_record_dict))
        return table

    def run(self, user_range, mode='-s'):
        assert mode in ('-s', '-p', '-e')
        ids = select_experiments(user_range, self.exp_record_dict)
        if len(ids)>1 and mode == '-p':
            import multiprocessing
            p = multiprocessing.Pool(processes=multiprocessing.cpu_count())
            p.map(partial(run_experiment_ignoring_errors, **self.run_args), ids)
        else:
            for experiment_identifier in ids:
                load_experiment(experiment_identifier).run(raise_exceptions=mode=='-e', display_results = False, **self.run_args )
        _warn_with_prompt('Finished running {} experiment{}.'.format(len(ids), '' if len(ids)==1 else 's'), use_prompt=not self.close_after)

    def test(self, user_range):
        ids = select_experiments(user_range, self.exp_record_dict)
        for experiment_identifier in ids:
            load_experiment(experiment_identifier).test()

    def help(self):
        _warn_with_prompt(self.HELP_TEXT, prompt = 'Press Enter to exit help.', use_prompt=not self.close_after)

    def show(self, user_range):
        for rid in select_experiment_records(user_range, self.exp_record_dict, flat=True):
            load_experiment_record(rid).show()
        _warn_with_prompt(use_prompt=not self.close_after)

    def results(self, user_range = 'all'):
        record_ids = select_experiment_records(user_range, self.exp_record_dict)
        with IndentPrint("Results:", show_line=True, show_end=True):
            for erid in record_ids:
                record = load_experiment_record(erid)
                with IndentPrint(erid, show_line=True, show_end=True):
                    try:
                        result = record.get_result()
                        record.get_experiment().display_last(result, err_if_none=False)
                    except NoSavedResultError:
                        print '<No result was saved>'
                    except Exception as err:
                        print err

        _warn_with_prompt(use_prompt=not self.close_after)

    def errortrace(self, user_range):
        record_ids = select_experiment_records(user_range, self.exp_record_dict, flat=True)
        with IndentPrint("Error Traces:", show_line=True, show_end=True):
            for erid in record_ids:
                with IndentPrint(erid, show_line=True):
                    record = load_experiment_record(erid)
                    print record.get_error_trace()
        _warn_with_prompt(use_prompt=not self.close_after)

    def delete(self, user_range):
        record_ids = select_experiment_records(user_range, self.exp_record_dict, flat=True)
        print '{} out of {} Records will be deleted.'.format(len(record_ids), sum(len(recs) for recs in self.exp_record_dict.values()))
        with IndentPrint():
            print ExperimentRecordBrowser.get_record_table(record_ids, )
        response = raw_input('Type "yes" to continue. >')
        if response.lower() == 'yes':
            clear_experiment_records(record_ids)
            print 'Records deleted.'
        else:
            _warn_with_prompt('Records were not deleted.', use_prompt=not self.close_after)

    def call(self, user_range):
        ids = select_experiments(user_range, self.exp_record_dict)
        for experiment_identifier in ids:
            load_experiment(experiment_identifier)()

    def select(self, user_range):
        record_ids = select_experiment_records(user_range, self.exp_record_dict, flat=True)
        with IndentPrint():
            print ExperimentRecordBrowser.get_record_table(record_ids)
        _warn_with_prompt('Selection "{}" includes {} out of {} records.'.format(user_range, len(record_ids), sum(len(recs) for recs in self.exp_record_dict.values())), use_prompt=not self.close_after)

    def side_by_side(self, user_range):
        record_ids = select_experiment_records(user_range, self.exp_record_dict, flat=True)
        records = [ExperimentRecord.from_identifier(rid) for rid in record_ids]
        texts = ['{title}\n{sep}\n{info}\n{sep}\n{output}\n{sep}'.format(title=rid, sep='='*len(rid), info=record.info.get_text(), output=record.get_log())
                 for rid, record in zip(record_ids, records)]
        print side_by_side(texts, max_linewidth=128)
        _warn_with_prompt(use_prompt=not self.close_after)

    def compare(self, user_range):
        record_ids = select_experiment_records(user_range, self.exp_record_dict, flat=True)
        compare_experiment_records(record_ids)
        _warn_with_prompt(use_prompt=not self.close_after)

    def records(self, ):
        browse_experiment_records(self.exp_record_dict.keys())

    def allruns(self, ):
        self.just_last_record = not self.just_last_record

    def filter(self, user_range):
        self._filter = user_range if user_range!='--clear' else None

    def view(self, mode):
        self.view_mode = mode

    def compare_results(self, user_range):
        experiment_ids = select_experiments(user_range, self.exp_record_dict)
        load_experiment(experiment_ids[0]).compare_results(experiment_ids)
        _warn_with_prompt(use_prompt=not self.close_after)

    def explist(self, surround = ""):
        print "\n".join([surround+k+surround for k in self.exp_record_dict.keys()])
        _warn_with_prompt(use_prompt=not self.close_after)

    def quit(self):
        return self.QUIT


def select_experiments(user_range, exp_record_dict):

    experiment_list = exp_record_dict.keys()

    number_range = interpret_numbers(user_range)
    if number_range is not None:
        return [experiment_list[i] for i in number_range]
    elif user_range == 'all':
        return experiment_list
    elif user_range.startswith('has:'):
        phrase = user_range[4:]
        return [exp_id for exp_id in experiment_list if phrase in exp_id]
    elif user_range.startswith('hasnot:'):
        phrase = user_range[len('hasnot:'):]
        return [exp_id for exp_id in experiment_list if phrase not in exp_id]
    elif user_range in ('unfinished', 'invalid'):  # Return all experiments where all records are unfinished/invalid
        record_filters = select_experiment_records(user_range, exp_record_dict)
        return [exp_id for exp_id in experiment_list if all(record_filters[exp_id])]
    else:
        raise Exception("Don't know how to use input '{}' to select experiments".format(user_range))


def select_experiment_records(user_range, exp_record_dict, flat=True):
    """
    :param user_range:
    :param exp_record_dict: An OrderedDict<experiment_name: list<experiment_record_name>>
    :param flat: Return a list of experiment records, instead of an OrderedDict
    :return: if not flat, an An OrderedDict<experiment_name: list<experiment_record_name>>
        otherwise a list<experiment_record_name>
    """
    filters = _filter_records(user_range, exp_record_dict)
    filtered_dict = OrderedDict((k, [v for v, f in izip_equal(exp_record_dict[k], filters[k]) if f]) for k in exp_record_dict.keys())
    if flat:
        return [record_id for records in filtered_dict.values() for record_id in records]
    else:
        return filtered_dict


def _filter_records(user_range, exp_record_dict):
    """
    :param user_range:
    :param exp_record_dict:
    :return:
    """

    def _bitwise(op, filter_set_1, filter_set_2):
        assert op in ('and', 'or')
        filter_set_3 = filter_set_1.copy()
        for k in filter_set_1.keys():
            filter_set_3[k] = [(a or b) if op=='or' else (a and b) for a, b in izip_equal(filter_set_1[k], filter_set_2[k])]
        return filter_set_3

    if '|' in user_range:
        return reduce(lambda a, b: _bitwise('or', a, b), [_filter_records(subrange, exp_record_dict) for subrange in user_range.split('|')])
    if '&' in user_range:
        return reduce(lambda a, b: _bitwise('and', a, b), [_filter_records(subrange, exp_record_dict) for subrange in user_range.split('&')])
    base = OrderedDict((k, [False]*len(v)) for k, v in exp_record_dict.iteritems())
    number_range = interpret_numbers(user_range)
    keys = exp_record_dict.keys()
    if number_range is not None:
        for i in number_range:
            base[keys[i]] = [True]*len(base[keys[i]])
    elif '.' in user_range:
        exp_rec_pairs = interpret_record_identifier(user_range)
        for exp_number, rec_number in exp_rec_pairs:
            base[keys[exp_number]][rec_number] = True
    elif user_range == 'old':
        for k, v in base.iteritems():
            base[k] = ([True]*(len(v)-1)+[False]) if len(v)>0 else []
    elif user_range == 'unfinished':
        for k, v in base.iteritems():
            base[k] = [load_experiment_record(rec_id).info.get_field(ExpInfoFields.STATUS) != ExpStatusOptions.FINISHED for rec_id in exp_record_dict[k]]
        # filtered_dict = OrderedDict((exp_id, [rec_id for rec_id in records if load_experiment_record(rec_id).info.get_field(ExpInfoFields.STATUS) != ExpStatusOptions.FINISHED]) for exp_id, records in exp_record_dict.iteritems())
    elif user_range == 'invalid':
        for k, v in base.iteritems():
            base[k] = [load_experiment_record(rec_id).is_valid() is False for rec_id in exp_record_dict[k]]
    elif user_range == 'all':
        for k, v in base.iteritems():
            base[k] = [True]*len(v)
    elif user_range == 'errors':
        for k, v in base.iteritems():
            base[k] = [load_experiment_record(rec_id).info.get_field(ExpInfoFields.STATUS)==ExpStatusOptions.ERROR for rec_id in exp_record_dict[k]]
    else:
        raise Exception("Don't know how to interpret subset '{}'".format(user_range))
    return base


def interpret_numbers(user_range):
    """
    :param user_range: A string specifying a range of numbers.  Eg.
        interpret_numbers('4-6')==[4,5,6]
        interpret_numbers('4,6')==[4,6]
        interpret_numbers('4,6-9')==[4,6,7,8,9]
    :return: A list of integers, or None if the input is not numberic
    """
    if all(d in '0123456789-,' for d in user_range):
        numbers_and_ranges = user_range.split(',')
        numbers = [n for lst in [[int(s)] if '-' not in s else range(int(s[:s.index('-')]), int(s[s.index('-')+1:])+1) for s in numbers_and_ranges] for n in lst]
        return numbers
    else:
        return None


def interpret_record_identifier(user_range):
    """
    You can identify a single record with, eg 3.4, meaning "record 4 from experiment 3:.
    You can identify a range with, eg 3.1-3, meaning "records 3.1, 3.2, 3.3"
    :param user_range: The user input
    :return: A list of 2 tuples (exp_no, record_no).  e.g [(3, 4)], or [(3, 1), (3, 2), (3, 3)] in the above examples.
    """
    if ',' in user_range:
        parts = user_range.split(',')
        return [pair for p in parts for pair in interpret_record_identifier(p)]
    if '.' not in user_range:
        return None
    else:
        exp_number, record_numbers = user_range.split('.')
        return [(int(exp_number), rec_num) for rec_num in interpret_numbers(record_numbers)]


def compare_experiment_records(record_identifiers):

    records = [load_experiment_record(ident) for ident in record_identifiers]
    # info_results = OrderedDict([(identifier, record.get_info()) for identifier, record in zip(record_identifiers, records)]])

    funtion_names = [record.info.get_field(ExpInfoFields.FUNCTION) for record in records]
    args = [record.info.get_field(ExpInfoFields.ARGS) for record in records]
    results = [record.get_result() for record in records]

    common_args, different_args = separate_common_items(args)

    def lookup_fcn(identifier, column):
        index = record_identifiers.index(identifier)
        if column=='Function':
            return funtion_names[index]
        elif column=='Run Time':
            return records[index].info.get_field_text(ExpInfoFields.RUNTIME)
        elif column=='Common Args':
            return ', '.join('{}={}'.format(k, v) for k, v in common_args)
        elif column=='Different Args':
            return ', '.join('{}={}'.format(k, v) for k, v in different_args[index])
        elif column=='Result':
            return results[index]
        else:
            bad_value(column)

    rows = build_table(lookup_fcn,
        row_categories=record_identifiers,
        column_categories=['Function', 'Run Time', 'Common Args', 'Different Args', 'Result'],
        prettify_labels=False
        )

    print tabulate(rows)


def browse_experiment_records(names = None, filter_text = None):
    """
    Browse through experiment records.

    :param names: Filter by names of experiments
    :param filter_text: Filter by regular expression
    :return:
    """

    experiment_record_browser = ExperimentRecordBrowser(experiment_names=names, filter_text=filter_text)
    experiment_record_browser.launch()


def select_experiment_records_from_list(user_range, experiment_records):
    return [rec_id for rec_id, f in izip_equal(experiment_records, _filter_experiment_record_list(user_range, experiment_records)) if f]


def _filter_experiment_record_list(user_range, experiment_record_ids):
    if user_range=='all':
        return [True]*len(experiment_record_ids)
    elif user_range=='new':
        return detect_duplicates(experiment_record_ids, key=record_id_to_experiment_id, keep_last=True)
        # return [n for n, is_old in izip_equal(get_record_ids(), old) if not old]
    elif user_range=='old':
        return [not x for x in _filter_records(user_range, 'new')]
    elif user_range=='orphans':
        orphans = []
        for i, record_id in enumerate(experiment_record_ids):
            info = load_experiment_record(record_id).info
            if 'Module' in info:
                try:
                    import_module(info['Module'])
                    if not record_id_to_experiment_id(record_id) in GLOBAL_EXPERIMENT_LIBRARY:
                        orphans.append(True)
                    else:
                        orphans.append(False)
                except ImportError:
                    orphans.append(True)
            else:  # They must be old... lets kill them!
                orphans.append(True)
        return orphans
    else:
        which_ones = interpret_numbers(user_range)
        if which_ones is None:
            raise Exception('Could not interpret user range: "{}"'.format(user_range))
        filters = [False]*len(experiment_record_ids)
        for i in which_ones:
            filters[i] = True
        return filters



class ExperimentRecordBrowser(object):

    QUIT = 'Quit'
    HELP_TEXT = """
    q:                  Quit
    r:                  Refresh
    filter <text>       filter experiments
    viewfilters         View all filters on these results
    showall:            Show all experiments ever
    allnames:           Remove any name filters
    show <number>       Show experiment with number
    side_by_side 4,6,9       Compare experiments by their numbers.
    clearall            Delete all experements from your computer
"""

    def __init__(self, experiment_names = None, filter_text = None, raise_display_errors=False):
        """
        Browse through experiment records.

        :param names: Filter by names of experiments
        :param filter_text: Filter by regular expression
        :return:
        """
        self.experiment_names = experiment_names
        self.filters = [filter_text]
        self.record_ids = self.reload_ids()
        self.raise_display_errors = raise_display_errors

    def reload_ids(self):
        return get_all_record_ids(experiment_ids= self.experiment_names, filters=self.filters)

    @staticmethod
    def get_record_table(record_ids = None, headers = ('#', 'Identifier', 'Start Time', 'Duration', 'Status', 'Valid', 'Notes', 'Result'), raise_display_errors = False):

        d = {
            '#': lambda: i,
            'Identifier': lambda: record_id,
            'Start Time': lambda: experiment_record.info.get_field_text(ExpInfoFields.TIMESTAMP, replacement_if_none='?'),
            'Duration': lambda: experiment_record.info.get_field_text(ExpInfoFields.RUNTIME, replacement_if_none='?'),
            'Status': lambda: experiment_record.info.get_field_text(ExpInfoFields.STATUS, replacement_if_none='?'),
            'Args': lambda: experiment_record.info.get_field_text(ExpInfoFields.ARGS, replacement_if_none='?'),
            'Valid': lambda: experiment_record.get_invalid_arg_note(),
            'Notes': lambda: experiment_record.info.get_field_text(ExpInfoFields.NOTES, replacement_if_none='?'),
            'Result': lambda: experiment_record.get_one_liner(),
            }

        def get_col_info(headers):
            info = []
            for h in headers:
                try:
                    info.append(d[h]())
                except:
                    info.append('<Error displaying info>')
                    if raise_display_errors:
                        raise
            return info

        rows = []
        for i, record_id in enumerate(record_ids):
            experiment_record = load_experiment_record(record_id)
            rows.append(get_col_info(headers))
        assert all_equal([len(headers)]+[len(row) for row in rows]), 'Header length: {}, Row Lengths: \n {}'.format(len(headers), [len(row) for row in rows])
        return tabulate(rows, headers=headers)

    def launch(self):

        func_lookup = {
            'q': self.quit,
            'h': self.help,
            'filter': self.filter,
            'showall': self.showall,
            'args': self.args,
            'rmfilters': self.rmfilters,
            'viewfilters': self.viewfilters,
            'side_by_side': self.compare,
            'show': self.show,
            'search': self.search,
            'delete': self.delete,
        }

        while True:

            print "=============== Experiment Records =================="
            # print tabulate([[i]+e.get_row() for i, e in enumerate(record_infos)], headers=['#']+_ExperimentInfo.get_headers())
            print self.get_record_table(self.record_ids, raise_display_errors = self.raise_display_errors)
            print '-----------------------------------------------------'

            if self.experiment_names is not None or len(self.filters) != 0:
                print 'Not showing all experiments.  Type "showall" to see all experiments, or "viewfilters" to view current filters.'
            user_input = raw_input('Enter Command (show # to show and experiment, or h for help) >>')
            parts = shlex.split(user_input)
            if len(parts)==0:
                print "You need to specify an command.  Press h for help."
                continue
            cmd = parts[0]
            args = parts[1:]

            try:
                if cmd not in func_lookup:
                    raise _warn_with_prompt('Unknown Command: {}'.format(cmd))
                else:
                    return_val = func_lookup[cmd](*args)
                    if return_val==self.QUIT:
                        break
            except Exception as e:
                res = raw_input('%s: %s\nEnter "e" to view the stacktrace, or anything else to continue.' % (e.__class__.__name__, e.message))
                if res == 'e':
                    raise

    def _select_records(self, user_range):
        return select_experiment_records_from_list(user_range, self.record_ids)

    def quit(self):
        return self.QUIT

    def help(self):
        _warn_with_prompt(self.HELP_TEXT)

    def filter(self, filter_text):
        self.filters.append(filter_text)
        self.record_ids = self.reload_ids()

    def showall(self):
        self.filters = []
        self.experiment_names = None
        self.record_ids = self.reload_ids()

    def args(self, user_range):
        print self.get_record_table(self._select_records(user_range), headers=['Identifier', 'Args'])

    def rmfilters(self):
        self.filters = []
        self.record_ids = self.reload_ids()

    def viewfilters(self):
        _warn_with_prompt('Filtering for: \n  Names in {}\n  Expressions: {}'.format(self.experiment_names, self.filters))

    def compare(self, user_range):
        identifiers = self._select_records(user_range)
        compare_experiment_records(identifiers)
        _warn_with_prompt('')

    def show(self, user_range):
        for rid in self._select_records(user_range):
            load_experiment_record(rid).show()
        _warn_with_prompt('')

    def search(self, filter_text):
        print 'Found the following Records: '
        print self.get_record_table([rid for rid in self.record_ids if filter_text in rid])
        _warn_with_prompt()

    def delete(self, user_range):
        ids = self._select_records(user_range)
        print 'We will delete the following experiments:'
        print self.get_record_table(ids)
        conf = raw_input("Going to clear {} of {} experiment records shown above.  Enter 'yes' to confirm: ".format(len(ids), len(self.record_ids)))
        if conf=='yes':
            clear_experiment_records(ids=ids)
            assert not any(has_experiment_record(rid) for rid in ids), "Failed to delete them?"
        else:
            _warn_with_prompt("Did not delete experiments")
        self.record_ids = self.reload_ids()


if __name__ == '__main__':
    browse_experiment_records()



from artemis.general.newline_writer import TextWrappingPrinter

__author__ = 'peter'


def test_newline_writer():

    text = """
fdsafdsafdsa gfsdgbfd fd svfdsb vfdsvfds dsacdascdas  vfdsbfdsbfdsbfdsbfdsbd

gfdsgfdsgfds

gfdsgfdsgfdsgfdgfdsgfdsgfdsbfdsgfds
"""

    t = TextWrappingPrinter()
    t.write(text)


if __name__ == '__main__':

    test_newline_writer()



from contextlib import contextmanager
from matplotlib import pyplot as plt
__author__ = 'peter'

"""
This allows you to attach callbacks to pyplots show/draw functions.  You may, for instance, want to save a figure
whenever plt.show is called.  After importing this module, you will be able to do this with the following:

    from artemis.plotting.manage_plotting import ShowContext

    def my_show_fcn(fig):
        ... save figure or whatever you want to do

    with ShowContext(my_show_fcn, clear_others=False):
        ... everything called in here will result in your function being called before show.  Set clear_others to false
        to remove the existing action of plt.show() (which displays the figure).

Alternatively, if you wish to globally set a callback and never change it, you can go:

    from artemis.plotting.manage_plotting import set_show_callback

    set_show_callback(my_show_fcn)


You can also do the same for drawing figures (showing without hanging on the figure) using DrawContext, set_draw_callback.
"""

_ORIGINAL_SHOW_CALLBACK = plt.show
_ORIGINAL_PLT_DRAW = plt.draw


class FigureCallBackManager(object):

    def __init__(self, callbacks):
        self.callbacks = callbacks

    def __call__(self, fig=None, **kwargs):
        if fig is None:
            fig = plt.gcf()
        for cb in self.callbacks:
            cb(fig, **kwargs)

    def insert_callback(self, cb, idx = 0):
        old_callbacks = list(self.callbacks)
        assert cb not in self.callbacks, "You've included the same callback object already"
        self.callbacks.insert(idx, cb)
        return old_callbacks

    def set_callback(self, cb):
        """
        :param cb: Set a callback and remove all others.
        """
        old_callbacks = list(self.callbacks)
        self.callbacks = [cb]
        return old_callbacks

    def set_callbacks(self, callbacks):
        self.callbacks = callbacks

    def remove_callback(self, callbacks):
        self.callbacks.remove(callbacks)

    def clear_callbacks(self):
        self.callbacks = []


@contextmanager
def WhatToDoOnShow(state):
    """
    Deterine whether or not show blocks.
    *Note: In test mode, this is overridden, and show will never block.
    :param state: One of:
        'hang': Show and hang
        'draw': Show but keep on going
        False: Don't show figures
    :return:
    """
    assert state in ('hang', 'draw', False)

    def new_show(*args, **kwargs):

        if state == 'hang':
            _ORIGINAL_SHOW_CALLBACK(*args, **kwargs)
        elif state == 'draw':
            plt.draw()
            plt.pause(0.00001)
        elif state == False:
            pass

    with ShowContext(new_show, clear_others=True):
        yield


class ShowContext(object):

    def __init__(self, callback, clear_others = False):
        self.callback = callback
        self.clear_others = clear_others

    def __enter__(self):
        self.old = plt.show
        plt.show = self._show

    def __exit__(self, exc_type, exc_val, exc_tb):
        plt.show = self.old

    def _show(self, *args, **kwargs):
        if 'block' in kwargs and kwargs['block'] is False:
            # This is treated a special case.  We treat plt.show(block=False) as a separate function.
            # It gets called by plt.pause() so we could get an infinite loop if we didn't do this.
            _ORIGINAL_SHOW_CALLBACK(*args, **kwargs)
        else:
            self.callback(*args, **kwargs)
            if not self.clear_others:
                self.old(*args, **kwargs)


class DrawContext(object):
    # TODO: Integrate this with redraw_figure(fig)
    # It mainly just useful for saving updating plots in experiments.

    def __init__(self, callback, clear_others = False):
        self.callback = callback
        self.clear_others = clear_others

    def __enter__(self):
        self.old = plt.draw
        plt.draw = self.callback if self.clear_others else self._draw_with_others

    def __exit__(self, exc_type, exc_val, exc_tb):
        plt.draw = self.old

    def _draw_with_others(self, *args, **kwargs):
        self.old(*args, **kwargs)
        self.callback(*args, **kwargs)


def set_show_callback(callback):
    """
    Perminantly changes the action of plt.show().  WARNING: This function can lead to some really confusing bugs.  Only
    use if you really know what you're doing.
    :param callback:
    """
    plt.show = callback


def set_draw_callback(callback):
    """
    Perminantly changes the action of plt.draw().  WARNING: This function can lead to some really confusing bugs.  Only
    use if you really know what you're doing.
    :param callback:
    """
    plt.draw = callback



import logging
from collections import OrderedDict
import pickle
import os


class PersistentOrderedDict(OrderedDict):
    """
    A Persisten ordered dict.  Usage:

    with PersistentOrderedDict('my_file.pkl') as pod:
        pod['a'] = [1, 2, 3]
        pod['b'] = [4, 5, 6]

    with PersistentOrderedDict('my_file.pkl') as pod:
        assert pod.items() == [('a', [1, 2, 3]), ('b', [4, 5, 6])]

    This is similar to python's built in "shelve" module, but
    - It is ordered,
    - It is used in a "with" statement.
    """

    def __init__(self, file_path, pickle_protocol=2):
        self.file_path = file_path
        if os.path.exists(self.file_path):
            try:
                with open(self.file_path, 'rb') as f:
                    items = pickle.load(f)
            except:
                logging.critical("WARNING: Failed to unpickle file: {}.  Starting from scratch instead".format(self.file_path))
                items = []
        else:
            items = []
        self.pickle_protocol = pickle_protocol
        OrderedDict.__init__(self, items)

    def __enter__(self):
        return self

    def close(self):
        with open(self.file_path, 'w') as f:
            pickle.dump(self.items(), f, protocol=self.pickle_protocol)

    def __exit__(self, thing1, thing2, thing3):
        self.close()

    def get_data(self):
        return OrderedDict(self)


from logging import Logger
from time import time
from collections import OrderedDict

__author__ = 'peter'


class EZProfiler(object):

    def __init__(self, profiler_name = 'Profile', print_result = True, print_on_start=False, record_stop = True):
        """
        :param profiler_name: The name of this profiler (will be included in report)
        :param print_result: Print
        :param record_stop:
        :return:
        """
        self.print_result = print_result
        self.profiler_name = profiler_name
        self.record_stop = record_stop
        self.print_on_start = print_on_start
        self._lap_times = OrderedDict()
        self._lap_times['Start'] = time()

    def lap(self, lap_name = None):
        """
        :param lap_name: How to identify this name
        """
        assert lap_name not in ('Start', 'Stop'), "Names 'Start' and 'Stop' are reserved."
        assert lap_name not in self._lap_times, "You already have a chekpoint called '%s'" % (lap_name, )
        t = time()
        self._lap_times[lap_name] = t
        return t

    def get_current_time(self):
        return time() - self._lap_times['Start']

    def __enter__(self):
        start_time = time()
        self.start_time = start_time
        if self.print_on_start:
            print '{} Started'.format(self.profiler_name)
        return self

    def __exit__(self, *args):
        if self.record_stop:
            self._lap_times['Stop'] = time()
        if self.print_result is True:
            self.print_elapsed()
        elif isinstance(self.print_result, Logger):
            self.print_result.info(self.get_report())

    def print_elapsed(self):
        print self.get_report()

    def get_report(self):
        keys = self._lap_times.keys()
        if len(keys)==2:
            return '%s: Elapsed time is %.4gs' % (self.profiler_name, self._lap_times['Stop']-self._lap_times['Start'])
        else:
            deltas = OrderedDict((key, self._lap_times[key] - self._lap_times[last_key]) for last_key, key in zip(keys[:-1], keys[1:]))
            return self.profiler_name + '\n  '.join(['']+['%s: Elapsed time is %.4gs' % (key, val) for key, val in deltas.iteritems()] +
                (['Total: %.4gs' % (self._lap_times.values()[-1] - self._lap_times.values()[0])] if len(deltas)>1 else []))




from artemis.general.should_be_builtins import bad_value
from artemis.plotting.data_conversion import put_data_in_grid, data_to_image, RecordBuffer, UnlimitedRecordBuffer
import numpy as np

from bokeh.models import GridPlot
from bokeh.document import Document
from bokeh.client import push_session
from bokeh.plotting import Figure, figure
from bokeh.palettes import Spectral6



_SESSION_COUNTER = 0
_ALL_PLOTS = {}
_CURRENT_SESSION = None
_CURRENT_PLOT = None
_URL = "default"
_SUBPLOT_DICT = {}
_GRIDPLOT = None

def set_url(url):
    global _URL
    _URL = url


def get_name():
    global _SESSION_COUNTER
    _SESSION_COUNTER += 1
    return "Figure_" + str(_SESSION_COUNTER)


class FakeFigure():
    def __init__(self, session):
        self.session = session

    def clf(self):
        global _CURRENT_SESSION
        _CURRENT_SESSION.document.clear()
        global _CURRENT_PLOT
        _CURRENT_PLOT = None
        global _ALL_PLOTS
        _ALL_PLOTS = {}
        global _SESSION_COUNTER
        _SESSION_COUNTER = 0
        global _SUBPLOT_DICT
        _SUBPLOT_DICT = {}
        global _GRIDPLOT
        _GRIDPLOT = None


def figure(*args,**kwargs):
    return FakeFigure(_session(*args, **kwargs))

def _session(name=None):
    if name == None:
        name = get_name()
    doc = Document()
    session = push_session(document=doc, session_id=name, url=_URL)
    print ("Session id: %s" %(name))
    session.show()
    global _CURRENT_SESSION
    _CURRENT_SESSION = session
    return session

def _get_or_make_session():
    return figure() if _CURRENT_SESSION is None else _CURRENT_SESSION

def _plot(model = None, **kwargs):
    session = _get_or_make_session()
    # Build Gridplot:
    num = _SUBPLOT_DICT["num"]
    rows = _SUBPLOT_DICT["rows"]
    cols = _SUBPLOT_DICT["cols"]
    global _GRIDPLOT
    if num == 1:
        children = [[None for _ in xrange(_SUBPLOT_DICT["cols"])] for _ in xrange(_SUBPLOT_DICT["rows"])]
    else:
        children = _GRIDPLOT.children
    newplot = model(**kwargs)
    global _CURRENT_PLOT
    _CURRENT_PLOT = newplot
    global _ALL_PLOTS
    _ALL_PLOTS[newplot._id] = newplot
    row_coordinate = int(np.ceil(float(num) / cols) - 1)
    col_coordinate = num % cols -1 if num % cols != 0 else cols -1
    children[row_coordinate][col_coordinate] = newplot
    _GRIDPLOT = GridPlot()
    _GRIDPLOT.children = children
    _CURRENT_SESSION.document.clear()
    _GRIDPLOT._detach_document()
    print _GRIDPLOT
    _CURRENT_SESSION.document.add_root(_GRIDPLOT)
    # global _CURRENT_PLOT
    # _CURRENT_PLOT = newplot
    return newplot



    # plot = model(**kwargs)
    # global _CURRENT_SESSION
    # _CURRENT_SESSION.document.add_root(plot)

def _get_or_make_plot(model, **kwargs):
    session = _get_or_make_session()
    return _plot(model, **kwargs) if _CURRENT_PLOT is None else _CURRENT_PLOT


def make_plot(model, **kwargs):
    return _plot(model, **kwargs)

def subplot(rows, cols, num, **kwargs):

    global _SUBPLOT_DICT
    _SUBPLOT_DICT = {"rows":rows, "cols":cols, "num":num, "kwargs":kwargs}


def draw():
    pass

def show():
    pass

def isinteractive():
    pass

def interactive(bool):
    pass

def gca():
    return _CURRENT_PLOT

def title(s, *args, **kwargs):
    gca().title = s

def plot_data_format(*args, **kwargs):
    '''
    This function should take data as variable as possible and put it into a multi_line plot

    :param args:
    :param kwargs:
    :return:
    '''

    if isinstance(args[0], np.ndarray) and isinstance(args[1], np.ndarray):
        x_data, y_data = args[:2]
    else:
        x_data = np.arange(len(args[0]))
        y_data = args[0]
    # print (x_data)
    # print (y_data)

    if x_data.ndim == 1:
        x_data = np.expand_dims(x_data,0)
    if y_data.ndim == 1:
        y_data = np.expand_dims(y_data,0)

    if y_data.shape != x_data.shape:
        if y_data.T.shape == x_data.shape:
            x_data = x_data.T
        elif y_data.shape[0] % x_data.shape[0] == 0:
                x_data = np.repeat(x_data,y_data.shape[0],0)
        else:
            print ("x-axis data and y-axis data not correct for multi-line plot")
    assert y_data.shape == x_data.shape

    # Remove all nan values because updating the data dict can't deal with nan values.
    # y_data_list = [(y_data[i][~np.isnan(y_data[i])]).tolist() for i in xrange(y_data.shape[0])]
    # x_data_list = [(x_data[i][~np.isnan(y_data[i])]).tolist() for i in xrange(y_data.shape[0])]
    y_data = np.nan_to_num(y_data)

    return x_data.tolist(), y_data.tolist()
    # return x_data[~np.isnan(y_data)], y_data[~np.isnan(y_data)]

    # return gca().multi_line(xs = x_data.tolist(), ys = y_data.tolist(), color = Spectral6[:y_data.shape[0]], line_width = 2 )


def _glyph_already_on_document(glyph):
    global _ALL_PLOTS
    existing_renderers  = []
    for p in _ALL_PLOTS.values():
        existing_renderers = existing_renderers + p.renderers
    return glyph._id in [r._id for r in existing_renderers]

class LinePlot(object):
    def __init__(self, yscale = None, **kwargs):
        self._plots = None
        self.kw = kwargs

    def update(self, data):
        # print(data)
        x_data, y_data = plot_data_format(np.arange(-data.shape[0]+1, 1), data.T)

        if self._plots is None:
            ax = make_plot(Figure, **self.kw)
            self._plots = gca().multi_line(xs = x_data, ys = y_data, color = Spectral6[:len(y_data)], line_width = 2 )
            # self._plots = plot(np.arange(-data.shape[0]+1, 1), data.T, **self.kw)
        else:
            _get_or_make_plot(Figure, **self.kw)
            if not _glyph_already_on_document(self._plots):
                _CURRENT_PLOT.renderers.append(self._plots)
            self._plots.data_source.data["xs"] = x_data
            self._plots.data_source.data["ys"] = y_data
            print self._plots.data_source.data

class HistoryPlot(LinePlot):
    def __init__(self, **kwargs):
        LinePlot.__init__(self,**kwargs)
        self._buffer = UnlimitedRecordBuffer()

    def update(self,data):
        if not np.isscalar(data):
            data = data.flatten()
        buffer_data = self._buffer(data)
        LinePlot.update(self,buffer_data)



class MovingPointPlot(LinePlot):

    def __init__(self, buffer_len=100, **kwargs):
        LinePlot.__init__(self, **kwargs)
        self._buffer = RecordBuffer(buffer_len)

    def update(self, data):
        if not np.isscalar(data):
            data = data.flatten()
        buffer_data = self._buffer(data)
        LinePlot.update(self, buffer_data)



class ImagePlot(object):

    def __init__(self, interpolation = 'nearest', show_axes = False, clims = None, aspect = 'auto', cmap = 'gray'):
        self._plot = None
        self._interpolation = interpolation
        self._show_axes = show_axes
        self._clims = clims
        self._aspect = aspect
        self._cmap = cmap

    def update(self, data):

        if data.ndim == 1:
            data = data[None]

        clims = ((np.nanmin(data), np.nanmax(data)) if data.size != 0 else (0, 1)) if self._clims is None else self._clims

        plottable_data = put_data_in_grid(data, clims = clims, cmap = self._cmap) \
            if not (data.ndim == 2 or data.ndim == 3 and data.shape[2] == 3) else \
            data_to_image(data, clims = clims, cmap = self._cmap)

        if self._plot is None:
            self._plot = imshow(plottable_data, interpolation = self._interpolation, aspect = self._aspect, cmap = self._cmap)
            if not self._show_axes:
                # self._plot.axes.get_xaxis().set_visible(False)
                self._plot.axes.tick_params(labelbottom = 'off')
                self._plot.axes.get_yaxis().set_visible(False)
            # colorbar()

        else:
            self._plot.set_array(plottable_data)
        self._plot.axes.set_xlabel('%.2f - %.2f' % clims)
            # self._plot.axes.get_caxis



def get_plot_from_data(data, mode, **plot_preference_kwargs):

    return \
        get_live_plot_from_data(data, **plot_preference_kwargs) if mode == 'live' else \
        get_static_plot_from_data(data, **plot_preference_kwargs) if mode == 'static' else \
        ImagePlot(**plot_preference_kwargs) if mode == 'image' else \
        bad_value(mode, 'Unknown plot modee: %s' % (mode, ))


def get_live_plot_from_data(data, line_to_image_threshold = 8, cmap = 'gray'):

    if isinstance(data, basestring):
        return TextPlot()

    is_scalar = np.isscalar(data) or data.shape == ()
    if is_scalar:
        data = np.array(data)

    is_1d = not is_scalar and data.size == np.max(data.shape)
    few_values = data.size < line_to_image_threshold

    if is_scalar or is_1d and few_values:
        return MovingPointPlot()
    elif is_1d:
        return MovingImagePlot()
    elif data.ndim == 2 and data.shape[1]<line_to_image_threshold:
        return LinePlot()
    elif data.ndim in (2, 3, 4, 5):
        return ImagePlot(cmap=cmap)
    else:
        raise NotImplementedError('We have no way to plot data of shape %s.  Make one!' % (data.shape, ))


def get_static_plot_from_data(data, line_to_image_threshold=8, cmap = 'gray'):

    if isinstance(data, basestring):
        return TextPlot()

    is_scalar = np.isscalar(data) or data.shape == ()
    if is_scalar or data.size==1:
        return TextPlot()

    is_1d = not is_scalar and data.size == np.max(data.shape)
    if is_1d:
        n_unique = len(np.unique(data))
        if n_unique == 2:
            return ImagePlot(cmap=cmap)
        else:
            return LinePlot()
    elif data.ndim == 2 and data.shape[1] < line_to_image_threshold:
        return LinePlot()
    else:
        return ImagePlot(cmap=cmap)



from __future__ import print_function

import os
from ConfigParser import NoSectionError, NoOptionError

import paramiko
from artemis.config import get_artemis_config_value

from artemis.fileman.config_files import get_config_value
from artemis.remote.child_processes import get_ssh_connection


def check_config_file(ip_address,file_path=".artemisrc"):
    '''
    Makes sure all required fields are present in ~./artemisrc.
    Also performs test for the different options if applicable
    :param ip_address: The section to look for. Remote ip is assumed. Makes no sense for local ip.
    :return:
    '''
    mandatory_options = ["username","python"]
    artemisrc_path = os.path.expanduser("~/%s"%file_path)
    for option in mandatory_options:
        try:
            get_artemis_config_value(section=ip_address,option=option)
        except NoSectionError:
            print("Section %s could not be found in %s. Please provide it." %(ip_address, artemisrc_path))
            raise
        except NoOptionError:
            print("Section %s does not contain option %s. Please provide it in %s" %(ip_address, option, artemisrc_path))
            raise

    # optional_options = ["private_key"]
    try:
        private_key_path = get_artemis_config_value(section=ip_address,option="private_key")
        assert os.path.isfile(private_key_path), "The path to the private_key for %s you specified in %s is not valid. You provided %s" %(ip_address, artemisrc_path, private_key_path)
    except NoOptionError:
        pass
    # username & private key setup tests:
    try:
        get_ssh_connection(ip_address)
    except paramiko.ssh_exception.AuthenticationException as e:
        if "Authentication failed" in e.message:
            print("An AuthenticationException is being raised. Make sure you have your private key set up correctly")
        else:
            print("An AuthenticationException is being raised. Did you specify the correct username for %s in %s? You provided the username %s"% (ip_address, artemisrc_path, get_artemis_config_value(section=ip_address,option="username")))
        raise
    except paramiko.ssh_exception.SSHException:
        try:
            private_key_path = get_artemis_config_value(section=ip_address,option="private_key")
            print ("Something is wrong with the private_key you specified in %s for %s . You provided %s" % (artemisrc_path, ip_address, private_key_path))
            raise
        except NoOptionError:
            private_key_path = os.path.join(os.path.expanduser("~"),".ssh/id_rsa")
            print("You did not provide a private_key path in %s. The default path %s appears to be wrongly set up. "
                  "Please make sure you have correctly set up your private key for %s " %(artemisrc_path,private_key_path,ip_address))


    #python tests:
    python_path = get_artemis_config_value(section=ip_address,option="python")

    command = "python -c 'import os; print(os.path.isfile(os.path.expanduser(\"%s\")))'"%python_path
    ssh_conn = get_ssh_connection(ip_address)
    _,stdout,stderr = ssh_conn.exec_command(command)
    assert stdout.read().strip()=="True", "The provided path to the remote python installation on %s does not exist. You provided %s" %(ip_address, python_path)

    command = "%s -c 'print(\"Success\")'" % python_path
    _,stdout,stderr = ssh_conn.exec_command(command)
    err = stderr.read().strip()
    assert stdout.read().strip()=="Success" and not err, "The provided python path on %s does not seem to point to a python executable. " \
                                                         "You provided %s, which resulted in the following error on the remote machine: " %(ip_address, python_path, err)


def simple_rsync(local_path, remote_path, ip_address, verbose=False):
    '''
    This method synchronizes local_path and all subfolders with remote_path at the given address.
    This method executes a system rsync call. This is not a general wrapper for rsync. The call is blocking.
    :param local_path:
    :param remote_path: Assumed to be relative to the home dir
    :param ip_address:
    :return:
    '''
    options = "-ah"
    if verbose:
        options += "v"

    local_path = os.path.expanduser(local_path)
    username = get_artemis_config_value(section=ip_address, option="username")
    if remote_path.startswith("~"):
        remote_path = remote_path[1:]
    if remote_path.startswith(("/")):
        remote_path = remote_path[1:]
    # to_path = "%s@%s:/home/%s/%s" % (username, address, username, remote_path)
    to_path = "%s@%s:~/%s" % (username, ip_address, remote_path)
    return rsync(options, from_path=local_path, to_path=to_path)


def rsync(options, from_path, to_path):
    '''
    basic rsync wrapper
    :param options:
    :param from_path:
    :param to_path:
    :return:
    '''
    import subprocess
    print ("Starting: rsync %s %s %s" % (options, from_path, to_path))
    if not type(options) is list:
        options = [options]
    command = subprocess.Popen(["rsync"] + options + [from_path, to_path], stdout=subprocess.PIPE, stderr=subprocess.PIPE, bufsize=1)
    if "v" in options:
        while True:
            line = command.stdout.readline()
            if line != '':
                print (line.rstrip())
            else:
                break
    err = command.stderr.read().strip()
    if err:
        msg = "rsync received messages on stderr. This might indicate that the command failed or, if you transferred to a remote server," \
              " it might just be some message received by the remote server. \n" \
              "This is because rsync automatically forwards all messages by the remote server to stderr. \n" \
              "If you are confident that the call succeeded although stderr received messages, then catch the RuntimeError accordingly.\n " \
              "The messages received are: \n %s" % err
        raise RuntimeError(msg)
    print("rsync finished")
    return True



from functools import partial

from artemis.plotting.db_plotting import dbplot, hold_dbplots, set_dbplot_figure_size
import numpy as np
from artemis.plotting.matplotlib_backend import MovingPointPlot
import time

__author__ = 'peter'


def demo_dbplot(n_frames = 1000):
    """
    Demonstrates the various types of plots.

    The appropriate plot can usually be inferred from the first input data.  In cases where there are multple ways to
    display the input data, you can use the plot_type argument.
    """
    # Approximate frame rates:
    # Macbook Air, MacOSX backend, mode=safe ~2.4 FPS
    # Macbook Air, MacOSX backend, mode=fast:  FAILS
    # Macbook Air, TkAgg backend, mode=safe: ~1.48 FPS
    # Macbook Air, TkAgg backend, mode=fast: ~4.4 FPS
    # Linux box, Qt4Agg backend, mode=safe: ~2.5 FPS
    # Linux box, Qt4Agg backend, mode=fast: (plot does not update)

    # set_dbplot_figure_size(150, 100)
    for i in xrange(n_frames):
        t_start = time.time()
        with hold_dbplots():  # Sets it so that all plots update at once (rather than redrawing on each call, which is slower)
            dbplot(np.random.randn(20, 20), 'Greyscale Image')
            dbplot(np.random.randn(20, 20, 3), 'Colour Image')
            dbplot(np.random.randn(15, 20, 20), "Many Images")
            dbplot(np.random.randn(3, 6, 20, 20, 3), "Colour Image Grid")
            dbplot([np.random.randn(15, 20, 3), np.random.randn(10, 10, 3), np.random.randn(10, 30, 3)], "Differently Sized images")
            dbplot(np.random.randn(20, 2), 'Two Lines')
            dbplot((np.linspace(-5, 5, 100)+np.sin(np.linspace(-5, 5, 100)*2), np.linspace(-5, 5, 100)), '(X,Y) Lines')
            dbplot([np.sin(i/20.), np.sin(i/15.)], 'Moving Point History')
            dbplot([np.sin(i/20.), np.sin(i/15.)], 'Bounded memory history', plot_type=partial(MovingPointPlot, buffer_len=10))
            dbplot((np.cos(i/10.), np.sin(i/11.)), '(X,Y) Moving Point History')
            dbplot((np.cos(i/np.array([10., 15., 20.])), np.sin(i/np.array([11., 16., 21.])))+np.array([0, 1, 2]), 'multi (X,Y) Moving Point History', plot_type='trajectory')
            dbplot(np.random.randn(20), 'Vector History')
            dbplot(np.random.randn(50), 'Histogram', plot_type = 'histogram')
            dbplot(np.random.randn(50), 'Cumulative Histogram', plot_type = 'cumhist')
            dbplot(('Veni', 'Vidi', 'Vici')[i%3], 'text-history')
            dbplot(('Veni', 'Vidi', 'Vici')[i%3], 'text-notice', plot_type='notice')
        if i % 10 == 0:
            print 'Frame Rate: {:3g}FPS'.format(1./(time.time() - t_start))


def demo_debug_dbplot():

    import pdb
    for i in xrange(1000):
        dbplot(np.random.randn(50, 2), 'a')
        print 'aaa'
        pdb.set_trace()
        dbplot(np.random.randn(10, 10), 'b')
        print 'bbb'
        pdb.set_trace()


if __name__ == '__main__':
    demo_dbplot()
    # demo_debug_dbplot()
    print("#"*40)
    raw_input("Terminate")


from artemis.fileman.smart_io import smart_load, smart_save
import numpy as np


def test_smart_image_io(plot = False):

    image = smart_load('https://raw.githubusercontent.com/petered/data/master/images/artemis.jpeg', use_cache=True)
    smart_save(image[:, ::-1, :], 'output/simetra.png')
    rev_image = smart_load('output/simetra.png')
    assert np.array_equal(rev_image, rev_image)
    if plot:
        from artemis.plotting.db_plotting import dbplot
        dbplot(image, 'Artemis')
        dbplot(rev_image, 'Simetra', hang=True)


if __name__ == '__main__':
    test_smart_image_io(plot=False)


from artemis.config import get_artemis_config_value

__author__ = 'peter'

_PLOTTING_SERVER = get_artemis_config_value(section='plotting', option='plotting_server', default_generator="")
_USE_SERVER = _PLOTTING_SERVER != ""

BACKEND = get_artemis_config_value(section='plotting', option='backend')
assert BACKEND in ('matplotlib', 'matplotlib-web', 'bokeh'), 'Your config file ~/.artimisrc lists "%s" as the backend.  Valid backends are "matplotlib" and "bokeh".  Change the file.' % (BACKEND, )

if BACKEND in ('matplotlib', 'matplotlib-web'):
    from matplotlib.pyplot import *
    from artemis.plotting.matplotlib_backend import *
elif BACKEND == 'bokeh':
    from artemis.plotting.bokeh_backend import *


def is_server_plotting_on():
    return _USE_SERVER


def set_server_plotting(state):
    global _USE_SERVER
    _USE_SERVER = state


def get_plotting_server_address():
    return _PLOTTING_SERVER



from artemis.config import get_artemis_config_value
from artemis.general.test_mode import set_test_mode
from artemis.plotting.db_plotting import dbplot
from artemis.plotting.web_backend import setup_web_plotting
import numpy as np
from matplotlib import pyplot as plt


__author__ = 'peter'


def test_plotting_server():

    if get_artemis_config_value(section='plotting', option='backend') != 'matplotlib-web':
        setup_web_plotting()

    for i in xrange(5):
        dbplot(np.random.randn(10, 10, 3), 'noise')
        dbplot(np.random.randn(20, 2), 'lines')
        plt.pause(0.1)


if __name__ == '__main__':
    set_test_mode(True)
    test_plotting_server()



import time
import warnings

import itertools
import matplotlib.pyplot as plt
import numpy as np
from artemis.experiments.experiment_record import \
    experiment_id_to_latest_record_id, get_experiment_info, load_experiment_record, ExperimentRecord, record_experiment, \
    delete_experiment_with_id, get_current_experiment_dir, experiment_function, open_in_experiment_dir, \
    experiment_testing_context
from artemis.experiments.deprecated import start_experiment, end_current_experiment
from artemis.general.test_mode import set_test_mode

__author__ = 'peter'

"""
The experiment interface.  Currently, this can be used in many ways.  We'd like to
converge on just using the interface demonstrated in test_get_latest_identifier.  So
use that if you're looking for an example.
"""


@experiment_function
def experiment_test_function():

    with warnings.catch_warnings():  # This is just to stop that stupid matplotlib warning from screwing up our logs.
        warnings.simplefilter('ignore')

        print 'aaa'
        plt.figure('sensible defaults')
        dat = np.random.randn(4, 5)
        plt.subplot(211)
        plt.imshow(dat)
        plt.subplot(212)
        plt.imshow(dat, interpolation = 'nearest', cmap = 'gray')
        plt.show()
        print 'bbb'
        plt.figure()
        plt.plot(np.random.randn(10))
        plt.show()


def assert_experiment_record_is_correct(exp_rec, show_figures=False):

    assert exp_rec.get_log() == 'aaa\nbbb\n'
    assert len(exp_rec.get_figure_locs()) == 2

    # Now assert that you can load an experiment from file and again display the figures.
    exp_dir = exp_rec.get_dir()
    exp_rec_copy = ExperimentRecord(exp_dir)
    assert exp_rec_copy.get_log() == 'aaa\nbbb\n'
    if show_figures:
        exp_rec_copy.show_figures()
    assert len(exp_rec_copy.get_figure_locs()) == 2


def test_experiment_with():
    """
    DEPRECATED INTERFACE

    This syntax uses the record_experiment function directly instead of hiding it.
    """

    with experiment_testing_context():
        delete_experiment_with_id('test_exp')
        with record_experiment(identifier = 'test_exp', print_to_console=True) as exp_rec:
            experiment_test_function()
            assert_experiment_record_is_correct(exp_rec, show_figures=False)


def test_start_experiment():
    """
    DEPRECATED INTERFACE

    An alternative syntax to the with statement - less tidy but possibly better
    for notebooks and such because it avoids you having to indent all code in the
    experiment.
    """

    with experiment_testing_context():
        record = start_experiment('start_stop_test')
        experiment_test_function()
        end_current_experiment()
        assert_experiment_record_is_correct(record, show_figures=False)


def test_run_and_show():
    """

    This is nice because it no longer required that an experiment be run and shown in a
    single session - each experiment just has a unique identifier that can be used to show
    its results whenevs.
    """
    with experiment_testing_context():
        experiment_record = experiment_test_function.run()
        assert_experiment_record_is_correct(experiment_record, show_figures=False)
        experiment_record.show()


def test_get_latest():
    with experiment_testing_context():
        record_1 = experiment_test_function.run()
        time.sleep(0.01)
        record_2 = experiment_test_function.run()
        identifier = experiment_id_to_latest_record_id('experiment_test_function')
        assert identifier == record_2.get_identifier()


def test_get_latest_identifier():

    with experiment_testing_context():
        exp_rec = experiment_test_function.run()
        print get_experiment_info('experiment_test_function')
        assert_experiment_record_is_correct(exp_rec)
        last_experiment_identifier = experiment_id_to_latest_record_id('experiment_test_function')
        assert last_experiment_identifier is not None, 'Experiment was run, this should not be none'
        same_exp_rec = load_experiment_record(last_experiment_identifier)
        assert_experiment_record_is_correct(same_exp_rec)


def test_accessing_experiment_dir():

    with experiment_testing_context():

        @experiment_function
        def access_dir_test():
            print '123'
            print 'abc'
            dir = get_current_experiment_dir()
            with open_in_experiment_dir('my_test_file.txt', 'w') as f:
                f.write('Experiment Directory is: {}'.format(dir))

        record = access_dir_test.run()

        filepaths = record.list_files()

        assert 'my_test_file.txt' in filepaths

        with record.open_file('my_test_file.txt') as f:
            assert f.read() == 'Experiment Directory is: {}'.format(record.get_dir())

        with record.open_file('output.txt') as f:
            assert f.read() == '123\nabc\n'


@experiment_function
def add_some_numbers_test_experiment(a=1, b=1):
    c = a + b
    print c
    return c


def test_saving_result():
    # Run root experiment
    with experiment_testing_context():
        rec = add_some_numbers_test_experiment.run()
        assert rec.get_result() == 2


def test_variants():

    @experiment_function
    def add_some_numbers(a=1, b=1):
        c=a+b
        print c
        return c

    with experiment_testing_context():

        # Create a named variant
        e1=add_some_numbers.add_variant('b is 3', b=3)
        assert e1.run().get_result()==4

        # Creata a sub-variant
        e11 = e1.add_variant('a is 2', a=2)
        assert e11.run().get_result() == 5

        # Create unnamed variant
        e2=add_some_numbers.add_variant(b=4)
        assert e2.run().get_result()==5
        assert e2.get_name() == 'add_some_numbers.b=4'

        # Create array of variants
        e_list = [add_some_numbers.add_variant(b=i) for i in xrange(5, 8)]
        assert [e.get_name() for e in e_list] == ['add_some_numbers.b=5', 'add_some_numbers.b=6', 'add_some_numbers.b=7']
        assert [e.run().get_result()==j for e, j in zip(e_list, range(6, 11))]

        # Create grid of variants
        e_grid = [add_some_numbers.add_variant(a=a, b=b) for a, b in itertools.product([2, 3], [4, 5, 6])]
        assert [e.get_name() for e in e_grid] == ['add_some_numbers.a=2,b=4', 'add_some_numbers.a=2,b=5', 'add_some_numbers.a=2,b=6',
                                                  'add_some_numbers.a=3,b=4', 'add_some_numbers.a=3,b=5', 'add_some_numbers.a=3,b=6']
        assert add_some_numbers.get_unnamed_variant(a=2, b=4).run().get_result()==6
        assert add_some_numbers.get_unnamed_variant(a=3, b=5).run().get_result()==8

        experiments = add_some_numbers.get_all_variants(include_roots=True, include_self=True)
        assert len(experiments)==13


if __name__ == '__main__':
    set_test_mode(True)
    test_get_latest_identifier()
    test_get_latest()
    test_run_and_show()
    test_experiment_with()
    test_start_experiment()
    test_accessing_experiment_dir()
    test_saving_result()
    test_variants()



import time
from artemis.general.test_mode import set_test_mode
from artemis.plotting.live_plotting import LiveStream, LivePlot, LiveCanal
from artemis.plotting.plotting_backend import MovingImagePlot, MovingPointPlot, LinePlot, ImagePlot, HistogramPlot
from itertools import count

__author__ = 'peter'
import numpy as np


def test_streaming(duration = 10):

    c = count()

    stream = LiveStream(lambda: {
        'text': ['Veni', 'Vidi', 'Vici'][c.next() % 3],
        'images': {
            'bw_image': np.random.randn(20, 20),
            'col_image': np.random.randn(20, 20, 3),
            'vector_of_bw_images': np.random.randn(11, 20, 20),
            'vector_of_colour_images': np.random.randn(11, 20, 20, 3),
            'matrix_of_bw_images': np.random.randn(5, 6, 20, 20),
            'matrix_of_colour_images': np.random.randn(5, 6, 20, 20, 3),
            },
        'line': np.random.randn(20),
        'lines': np.random.randn(20, 3),
        'moving_point': np.random.randn(),
        'moving_points': np.random.randn(3),
        })
    for i in xrange(duration):
        if i==1:
            start_time = time.time()
        elif i>1:
            print 'Average Frame Rate: %.2f FPS' % (i/(time.time()-start_time), )
        stream.update()


def test_dynamic_rebuild():

    def grab_data():
        if i < 10:
            data = {'bw_image': np.random.randn(20, 20)}
        else:
            data = {
                'bw_image': np.random.randn(20, 20),
                'lines': np.random.randn(2)
                }
        return data

    duration = 20
    stream = LiveStream(grab_data)
    for i in xrange(duration):
        if i==1:
            start_time = time.time()
        elif i>1:
            print 'Average Frame Rate: %.2f FPS' % (i/(time.time()-start_time), )
        stream.update()


def test_canaling(duration = 10):

    n_dims = 4

    # Don't be frightened by the double-lambda here - the point is just to get a callback
    # that spits out the same data when called in sequence.
    cb_constructor_1d = lambda: lambda rng = np.random.RandomState(0): rng.randn(n_dims)
    cb_image_data = lambda: lambda rng = np.random.RandomState(1): rng.rand(20, 30)
    cb_sinusoid_data = lambda: lambda c=count(): np.sin(c.next()/40.)

    canal = LiveCanal({
        'histo-mass': LivePlot(plot = HistogramPlot([-2.5, 0, 0.5, 1, 1.5, 2, 2.5], mode = 'mass'), cb = lambda: np.random.randn(np.random.randint(10))),
        'histo-density': LivePlot(plot = HistogramPlot([-2.5, 0, 0.5, 1, 1.5, 2, 2.5], mode = 'density'), cb = lambda: np.random.randn(np.random.randint(10))),
        '1d-default': cb_constructor_1d(),
        '1d-image': LivePlot(plot = MovingImagePlot(buffer_len=20), cb = cb_constructor_1d()),
        '1d-seizmic': LivePlot(plot = MovingPointPlot(), cb = cb_constructor_1d()),
        '1d-line': LivePlot(plot = LinePlot(), cb = cb_constructor_1d()),
        'image-autoscale': LivePlot(ImagePlot(), cb_image_data()),
        'image-overexposed': LivePlot(ImagePlot(clims = (0, 0.2)), cb_image_data()),
        'image-jet': LivePlot(ImagePlot(cmap='jet'), cb_image_data()),
        'trace-default': cb_sinusoid_data(),
        'trace-prescaled': LivePlot(MovingPointPlot(y_bounds=(-1, 1)), cb_sinusoid_data()),
        })

    for i in xrange(duration):
        canal.update()


if __name__ == '__main__':

    # set_test_mode(True)
    test_dynamic_rebuild()
    test_streaming(10)
    test_canaling(10)



from artemis.general.hashing import compute_fixed_hash
import numpy as np


def test_compute_fixed_hash():
    complex_obj = [1, 'd', {'a': 4, 'b': np.arange(10)}, (7, range(10))]
    assert compute_fixed_hash(complex_obj) == '285f0ff77c0cb6a5fc529ae55775f9dd'
    complex_obj[2]['b'][6]=0
    assert compute_fixed_hash(complex_obj) == '3124854bda236cd50f0cc04bd6f84935'
    complex_obj[2]['b'][6]=6  # Revert to old value
    assert compute_fixed_hash(complex_obj) == '285f0ff77c0cb6a5fc529ae55775f9dd'


if __name__ == '__main__':
    test_compute_fixed_hash()



from itertools import cycle

import matplotlib.pyplot as plt
from matplotlib.scale import ScaleBase, register_scale
from matplotlib.ticker import NullFormatter, AutoLocator, ScalarFormatter, NullLocator
from matplotlib.transforms import Transform
import numpy as np

from artemis.general.test_mode import is_test_mode
from artemis.ml.predictors.predictor_comparison import LearningCurveData


__author__ = 'peter'


def plot_learning_curves(learning_curves, xscale = 'sqrt', yscale = 'linear', hang = None, title = None, figure_name = None, y_title = 'Score'):
    """
    Plot a set of PredictionResults.  These can be obtained by running compare_predictors.
    See module test_compare_predictors for an example.

    :param learning_curves: An OrderedDict<str: LearningCurveData>
    :param xscale: {'linear', 'log', 'symlog', 'sqrt'}
    :param yscale: {'linear', 'log', 'symlog', 'sqrt'}
    :param hang: True for blocking plot.  False to keep executing.
    :param title: Title of the plot
    :return:
    """

    if isinstance(learning_curves, LearningCurveData):
        learning_curves = {'': learning_curves}

    colours = ['b', 'r', 'g', 'm', 'c', 'k']

    plt.figure(figure_name)

    legend = []

    for (record_name, record), colour in zip(learning_curves.iteritems(), cycle(colours)):
        times, scores = record.get_results()
        if np.array_equal(times.values()[0], [None]):  # Offline result... make a horizontal line
            assert all(len(s)==1 for s in scores.values())
            if 'Training' in scores:
                plt.axhline(scores['Training'], color=colour, linestyle = '--')
            if 'Test' in scores:
                plt.axhline(scores['Test'], color=colour, linestyle = '-')
        else:
            if 'Training' in scores:  # Online result... make a learning curve
                plt.plot(times['Training']+(1 if xscale == 'log' else 0), scores['Training'], '--'+colour)
            if 'Test' in scores:
                plt.plot(times['Test']+(1 if xscale == 'log' else 0), scores['Test'], '-'+colour)
        plt.gca().set_xscale(xscale)
        plt.gca().set_yscale(yscale)
        if 'Training' in scores:
            legend.append('%s-training' % record_name)
        if 'Test' in scores:
            legend.append('%s-test' % record_name)

    plt.xlabel('Epoch')
    plt.ylabel(y_title)
    plt.legend(legend, loc = 'best')
    if title is not None:
        plt.title(title)

    if hang is True:
        plt.ioff()
    elif hang is False or (hang is None and is_test_mode()):
        plt.ion()
    plt.show()


class SqrtScale(ScaleBase):
    """
    The Square-Root scale.  This is useful for displaying things like learning curves, which tend to have a lot of
    action early on and then flatten out later.  Unlike log-scale, it can represent zero.  The compression of larger
    values is also less dramatic than in log-scale.
    """

    name = 'sqrt'

    def __init__(self, *args, **kwargs):
        pass

    def get_transform(self):
        """
        Return a :class:`~matplotlib.transforms.Transform` instance representing the square-root transform
        """
        return SqrtTransform()

    def set_default_locators_and_formatters(self, axis):
        """
        Just took the code from LinearScale.set_default_locators_and_formatters
        """
        axis.set_major_locator(AutoLocator())
        axis.set_major_formatter(ScalarFormatter())
        axis.set_minor_locator(NullLocator())
        axis.set_minor_formatter(NullFormatter())


class SqrtTransform(Transform):
    """
    Mostly coopied from LogTransform
    """

    input_dims = 1
    output_dims = 1
    is_separable = True
    has_inverse = True

    def __init__(self):
        Transform.__init__(self)

    def transform_non_affine(self, a):
        return np.sqrt(np.abs(a))*np.sign(a)

    def inverted(self):
        return QuadTransform()


class QuadTransform(Transform):

    input_dims = 1
    output_dims = 1
    is_separable = True
    has_inverse = True

    def __init__(self):
        Transform.__init__(self)

    def transform_non_affine(self, a):
        return a**2*np.sign(a)

    def inverted(self):
        return SqrtTransform()


register_scale(SqrtScale)



from artemis.general.mymath import sigm
from artemis.ml.datasets.synthetic_clusters import get_synthetic_clusters_dataset
from artemis.ml.datasets.synthetic_linear import get_synthethic_linear_dataset
from artemis.ml.predictors.logistic_regressor import LogisticRegressor
from artemis.ml.predictors.train_and_test import assess_prediction_functions, mean_squared_error, percent_argmax_correct, \
    train_and_test_online_predictor, get_best_score
import numpy as np


def test_assess_prediction_functions(print_results=True):

    x_tr, y_tr, x_ts, y_ts = get_synthethic_linear_dataset(n_input_dims=80, n_output_dims=4).xyxy

    w = np.random.RandomState(1234).randn(80, 4)

    results = assess_prediction_functions(
        test_pairs=[('train', (x_tr, y_tr)), ('test', (x_ts, y_ts))],
        functions=[('prediction', lambda x: sigm(x.dot(w)))],
        costs = [mean_squared_error, percent_argmax_correct],
        print_results=print_results
        )

    assert results['train', 'prediction', 'percent_argmax_correct'] == 23.1


def test_train_online_predictor():
    """
    Demonstrates how to train the online predictor.
    """
    predictor = LogisticRegressor.from_init(n_in=20, n_out=4)
    info_score_pairs = train_and_test_online_predictor(
        dataset = get_synthetic_clusters_dataset(n_clusters=4, n_dims=20),
        predict_fcn=predictor.predict,
        train_fcn = predictor.train,
        score_measure = 'percent_argmax_correct',
        minibatch_size=10,
        test_epochs=[0, 0.2, 0.5, 0.8, 1],
        )
    assert [info.epoch for info, score in info_score_pairs] == [0, 0.2, 0.5, 0.8, 1]
    infos, scores = zip(*info_score_pairs)
    first_score, last_score = scores[0], scores[-1]
    assert first_score['test', 'predict', 'percent_argmax_correct'] < 40 and last_score['test', 'predict', 'percent_argmax_correct'] > 99
    info_score_pairs.get_best_value('test', 'predict', 'percent_argmax_correct') > 99


if __name__ == '__main__':
    test_assess_prediction_functions()
    test_train_online_predictor()



from collections import OrderedDict, Counter
import itertools
import contextlib

__author__ = 'peter'

all_equal = lambda *args: all(a == args[0] for a in args[1:])

def is_lambda(v):
    '''
    Source http://stackoverflow.com/a/3655857/2068168
    :param v:
    :return:
    '''
    LAMBDA = lambda:0
    return isinstance(v, type(LAMBDA)) and v.__name__ == LAMBDA.__name__

def bad_value(value, explanation = None):
    """
    :param value: Raise ValueError.  Useful when doing conditional assignment.
    e.g.
    dutch_hand = 'links' if eng_hand=='left' else 'rechts' if eng_hand=='right' else bad_value(eng_hand)
    """
    raise ValueError('Bad Value: %s%s' % (value, ': '+explanation if explanation is not None else ''))


def memoize(fcn):
    """
    Use this to decorate a function whose results you want to cache.
    """
    lookup = {}

    def memoization_wrapper(*args, **kwargs):
        # arg_signature = args + ('5243643254_kwargs_start_here', ) + tuple((k, kwargs[k]) for k in sorted(kwargs.keys()))
        hashable_arg_structure = arg_signature((args, kwargs))
        if hashable_arg_structure in lookup:
            return lookup[hashable_arg_structure]
        else:
            out = fcn(*args, **kwargs)
            lookup[hashable_arg_structure]=out
            return out

    memoization_wrapper.wrapped_fcn = fcn

    return memoization_wrapper


def arg_signature(arg):
    """
    Turn the argument into something hashable
    """
    if isinstance(arg, tuple):
        return tuple(arg_signature(a) for a in arg)
    elif isinstance(arg, list):
        return ('memoizationidentifier_list ',) + tuple(arg_signature(a) for a in arg)
    elif isinstance(arg, OrderedDict):
        return ('memoizationidentifier_ordereddict ',) + tuple((arg_signature(k), arg_signature(v)) for k, v in arg.iteritems())
    elif isinstance(arg, dict):
        return ('memoizationidentifier_dict ',) + tuple((arg_signature(k), arg_signature(arg[k])) for k in sorted(arg.keys()))
    else:
        return arg


def reducemap(func, sequence, initial=None, include_zeroth = False):
    """
    A version of reduce that also returns the intermediate values.
    :param func: A function of the form x_i_plus_1 = f(x_i, params_i)
        Where:
            x_i is the value passed through the reduce.
            params_i is the i'th element of sequence
            x_i_plus_i is the value that will be passed to the next step
    :param sequence: A list of parameters to feed at each step of the reduce.
    :param initial: Optionally, an initial value (else the first element of the sequence will be taken as the initial)
    :param include_zeroth: Include the initial value in the returned list.
    :return: A list of length: len(sequence), (or len(sequence)+1 if include_zeroth is True) containing the computed result of each iteration.
    """
    if initial is None:
        val = sequence[0]
        sequence = sequence[1:]
    else:
        val = initial
    results = [val] if include_zeroth else []
    for s in sequence:
        val = func(val, s)
        results.append(val)
    return results


def itermap(func, initial, n_steps=None, stop_func = None, include_zeroth = False):
    """
    Iterively call a function with the output of the previous call.
    :param func: A function of the form x_i_plus_1 = f(x_i)
    :param n_steps: The number of times to iterate
    :param initial: An initial value
    :param stop_func: Optionally, a function returning a boolean that, if true, causes the iteration to terminate (after the value has been added)
    :param include_zeroth: Include the initial value in the returned list.
    :return:  A list of length: n_steps, (or n_steps+1 if include_zeroth is True) containing the computed result of each iteration.
    """
    assert (n_steps is not None) or (stop_func is not None), 'You must either specify a number of steps or a stopping function.'
    val = initial
    results = [val] if include_zeroth else []
    for _ in (xrange(n_steps) if n_steps is not None else itertools.count(start=0, step=1)):
        val = func(val)
        results.append(val)
        if stop_func is not None and stop_func(val):
            break
    return results


def izip_equal(*iterables):
    """
    Zip and raise exception if lengths are not equal.

    Taken from solution by Martijn Pieters, here:
    http://stackoverflow.com/questions/32954486/zip-iterators-asserting-for-equal-length-in-python

    :param iterables:
    :return:
    """
    sentinel = object()
    for combo in itertools.izip_longest(*iterables, fillvalue=sentinel):
        if any(sentinel is c for c in combo):
            raise ValueError('Iterables have different lengths')
        yield combo


def remove_duplicates(sequence, hashable=True, key=None, keep_last=False):
    """
    Remove duplicates while maintaining order.
    :param sequence: The sequence you want to remove the duplicates in
    :param hashable: Set to True if your sequence contains unhashable items.  If your items are not hashable,
        you can set hashable=False.  The function will be slower, and test for dupicates based on equality rather than
        hash, which may in some cases give different results.
    :param key: Optionally, a function that extracts the identity from each element in sequence.  This allows you to be
        flexible in what you consider to be a "duplicate"
    :param keep_last: Keep the last element, rather than the first (only makes sense if key is not None)
    :returns: A list that maintains the order, but with duplicates removed
    """
    is_dup = detect_duplicates(sequence, hashable=hashable, key=key, keep_last=keep_last)
    return [x for x, is_duplicate in zip(sequence, is_dup) if not is_duplicate]


def detect_duplicates(sequence, hashable=True, key=None, keep_last=False):
    """
    Identify whether each element in a sequence is a duplicate of a previously existing element.

    Derived from solution by goes to Markus Jarderot from http://stackoverflow.com/a/480227/851699

    :param sequence: The sequence you want to remove the duplicates in
    :param hashable: Set to True if your sequence contains unhashable items.  If your items are not hashable,
        you can set hashable=False.  The function will be slower, and test for dupicates based on equality rather than
        hash, which may in some cases give different results.
    :param key: Optionally, a function that extracts the identity from each element in sequence.  This allows you to be
        flexible in what you consider to be a "duplicate"
    :param keep_last: Keep the last element, rather than the first.
    :returns: A list of booleans, which are True if the item is a duplicate and False otherwise.
    """
    if keep_last:
        sequence = sequence[::-1]
    if key is None:
        key = lambda x: x
    if hashable:
        seen = set()
        seen_add = seen.add
    else:
        seen = list()
        seen_add = seen.append
    if key is not None:
        sequence = [key(x) for x in sequence]
    is_dup = [(x in seen or seen_add(x) is 'This is such a hack') for x in sequence]
    if keep_last:
        is_dup = is_dup[::-1]
    return is_dup


def try_key(dictionary, key, default):
    """
    Try to get the value at dict[key]
    :param dictionary: A Python dict
    :param key: A key
    :param default: The value to return if the key doesn't exist
    :return: Either dictionary[key] or default if it doesn't exist.
    """
    try:
        return dictionary[key]
    except KeyError:
        return default


def separate_common_items(list_of_lists):
    """
    Given a list of lists of items, separate out the items that are common between the sublists
    into a single lists.

    :param list_of_lists: A list of lists of items
    :return: common_items, list_of_lists_of_different_items
        common_items: A list of items that are common to all the sublists.
        different_items: A list of lists of items that are not common between all sublists
    """
    are_dicts = all(isinstance(el, dict) for el in list_of_lists)
    if are_dicts:
        list_of_lists = [el.items() for el in list_of_lists]
    all_items = [item for list_of_items in list_of_lists for item in list_of_items]
    common_items = remove_duplicates([k for k, c in count_unique_items(all_items) if c==len(list_of_lists)], hashable=False)
    different_items = [[item for item in list_of_items if item not in common_items] for list_of_items in list_of_lists]
    if are_dicts:
        return dict(common_items), [dict(el) for el in different_items]
    else:
        return common_items, different_items


def count_unique_items(items):
    """
    Count the unique items in a list.  Similar to calling collections.Counter(items).items(), but it doesn't require
    that items be hashable.
    :return: A list<(item, item_count)>
    """
    unique_items = []
    unique_item_counts = []
    for item in items:
        if item in unique_items:
            unique_item_counts[unique_items.index(item)] +=1
        else:
            unique_items.append(item)
            unique_item_counts.append(1)
    return zip(unique_items, unique_item_counts)


def check(value, condition, string = ""):
    """
    Verify that the condition is true and return the value.

    Useful for conditional assignments, eg:
        xs = [x]*n_elements if not isinstance(x, (list, tuple)) else check(x, len(x)==n_elements)

    :param value:
    :param condition:
    :param string:
    :return:
    """
    assert condition, string
    return value


def remove_common_prefix(list_of_lists, max_elements=None):
    """
    Remove common elements starting each list in the list of lists.

        assert remove_common_prefix([[1, 2, 3, 4], [1, 2, 5], [1, 2, 3, 5]]) == [[3, 4], [5], [3, 5]]

    :param list_of_lists: A list of lists
    :param max_elements: Maximum number of elements to delete
    :return: Truncated list of lists
    """

    count = 0
    while max(len(parts) for parts in list_of_lists)>1:

        if max_elements is not None and count >= max_elements:
            break

        if all_equal(*[parts[0] for parts in list_of_lists]):
            list_of_lists = [parts[1:] for parts in list_of_lists]
        else:
            break
        count += 1
    return list_of_lists



from artemis.general.display import IndentPrint, CaptureStdOut, side_by_side

_desired = """
aaa
bbb
    ccc
    ddd
        eee
        fff
    ggg
    hhh
iii
jjj
"""


def test_indent_print():

    with CaptureStdOut() as cap:
        print 'aaa'
        print 'bbb'
        with IndentPrint():
            print 'ccc'
            print 'ddd'
            with IndentPrint():
                print 'eee'
                print 'fff'
            print 'ggg'
            print 'hhh'
        print 'iii'
        print 'jjj'

    assert '\n'+cap.read() == _desired


str1 = """For those who are led by the Spirit of God are the
children of God.
Romans 8:14"""

str2 = """But to you who are listening I say: Love your enemies,
do good to those who hate you, bless those who curse you, pray
for those who mistreat you.
Luke 6:27-28"""

desired = """For those who are led by the Spirit of God are the    But to you who are listening I say: Love your enemies,
children of God.                                      do good to those who hate you, bless those who curse you, pray
Romans 8:14                                           for those who mistreat you.
                                                      Luke 6:27-28
"""


def test_side_by_side():

    print 'String 1:\n{}'.format(str1)
    print 'String 2:\n{}'.format(str2)

    out = side_by_side([str1, str2])
    print 'Side by side:\n{}'.format(out)
    # assert out==desired  # Would work but pycharm automatically trims trailing spaces.

if __name__ == '__main__':
    # test_indent_print()
    test_side_by_side()



import pickle

from artemis.general.should_be_builtins import memoize
from artemis.ml.datasets.datasets import DataSet, DataCollection
from artemis.fileman.file_getter import get_file, unzip_gz
import numpy as np

__author__ = 'peter'


@memoize  # This should save time on tests and dataset should be immutable so it's all good.
def get_mnist_dataset(n_training_samples = None, n_test_samples = None, flat = False, join_train_and_val = False, binarize = False):
    """
    The MNIST DataSet - the Drosophila of machine learning.

    :param n_training_samples: Cap on the number of training samples
    :param n_test_samples: Cap on the number of test samples
    :param flat: Set to True if we just want flat 784-dimensional input data instead of 28x28 images.
    :param join_train_and_val: If true, merge the validation set into the training set.  (giving 60000 training examples,
        and 10000 test examples).  Otherwise you'll have 50000 training, 10000 validation, 10000 test.
    :param binarize: Binarize inputs by thresholding them at 0.5
    :return: A DataSet object containing the MNIST data
    """
    filename = get_file(
        relative_name = 'data/mnist.pkl',
        url = 'http://deeplearning.net/data/mnist/mnist.pkl.gz',
        data_transformation = unzip_gz)

    with open(filename) as f:
        data = pickle.load(f)

    x_tr, y_tr = data[0] if n_training_samples is None else (data[0][0][:n_training_samples], data[0][1][:n_training_samples])
    x_ts, y_ts = data[1] if n_test_samples is None else (data[1][0][:n_test_samples], data[1][1][:n_test_samples])
    x_vd, y_vd = data[2]
    if not flat:
        x_tr = x_tr.reshape(-1, 28, 28)
        x_ts = x_ts.reshape(-1, 28, 28)
        x_vd = x_vd.reshape(-1, 28, 28)
    if binarize:
        x_tr = x_tr>0.5
        x_ts = x_ts>0.5
        x_vd = x_vd>0.5

    return \
        DataSet(
            training_set=DataCollection(np.concatenate([x_tr, x_vd], axis=0), np.concatenate([y_tr, y_vd], axis=0)),
            test_set=DataCollection(x_ts, y_ts)
            ) \
        if join_train_and_val else \
        DataSet(
            training_set=DataCollection(x_tr, y_tr),
            test_set=DataCollection(x_ts, y_ts),
            validation_set=DataCollection(x_vd, y_vd)
            )
